#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2014 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.


from collections import defaultdict

from gi.repository import (
    Gdk,
    Gtk)

import lightdm_gtk_greeter_settings.helpers
from lightdm_gtk_greeter_settings.helpers import (
    check_path_accessibility,
    get_data_path,
    WidgetsEnum)

from gi.overrides import GLib


__all__ = ['MultiheadSetupDialog']

C_ = lambda t: lightdm_gtk_greeter_settings.helpers.C_('option|multihead', t)


class MultiheadSetupDialog(Gtk.Dialog):
    __gtype_name__ = 'MultiheadSetupDialog'

    def __new__(cls, monitors):
        builder = Gtk.Builder()
        builder.add_from_file(get_data_path('%s.ui' % cls.__gtype_name__))
        window = builder.get_object('multihead_setup_dialog')
        window.builder = builder
        window.monitors = monitors
        builder.connect_signals(window)
        window.init_window()
        return window

    class Widgets(WidgetsEnum):
        notebook = 'monitors_notebook'
        available_menu = 'available_menu'
        name = 'name_value'
        name_combo = 'name_combo'

        editor = 'editor_page'
        editor_add_button = 'editor_add_button'
        editor_add_menu_button = 'editor_add_menu_button'
        empty = 'empty_page'
        empty_add_button = 'empty_add_button'
        empty_add_menu_button = 'empty_add_menu_button'

    class PageData:
        group = None
        holder = None
        label = None
        ids = None
        # Entries
        name = None
        background = None

    builder = None
    monitors = None

    def init_window(self):
        self._widgets = self.Widgets(builder=self.builder)

        # To set size of widgets.notebook to size of widgets.editor
        self.realize()

        self._widgets.notebook.remove_page(self._widgets.notebook.page_num(self._widgets.editor))

        self._current_page = None
        self._defaults = {}
        self._page_to_data = {}

        screen = Gdk.Screen.get_default()
        self._available_monitors = [(screen.get_monitor_plug_name(i),
                                     Gtk.MenuItem(screen.get_monitor_plug_name(i)))
                                    for i in range(screen.get_n_monitors())]

        menu_header = Gtk.MenuItem(C_('Detected monitors:'))
        menu_header.set_sensitive(False)
        self._widgets.available_menu.append(menu_header)
        self._widgets.available_menu.append(Gtk.SeparatorMenuItem())
        self._widgets.available_menu.show_all()
        for name, item in self._available_monitors:
            self._widgets.available_menu.append(item)
            item.connect('activate', self.on_add_button_clicked, name)

    def run(self):
        editor_parent = self._widgets.editor.get_parent()
        if editor_parent:
            editor_parent.remove(self._widgets.editor)

        self._widgets.notebook.handler_block_by_func(self.on_switch_page)
        for page in self._widgets.notebook.get_children():
            if page != self._widgets.empty:
                self._widgets.notebook.remove_page(self._widgets.notebook.page_num(page))
        self._widgets.notebook.handler_unblock_by_func(self.on_switch_page)

        for data in self._page_to_data.values():
            for entry, ids in data.ids.items():
                for id_ in ids:
                    entry.disconnect(id_)

        self._page_to_data.clear()

        for group in self.monitors.groups:
            self._add_page(group)

        self._widgets.empty.props.visible = not self._page_to_data
        self._update_monitors_list()

        super().run()

    def _add_page(self, group):
        data = self.PageData()
        data.group = group

        data.holder = Gtk.Box()
        data.holder.show()
        data.label = Gtk.Label(group.entries['name'].value)

        close_button = Gtk.Button()
        close_button.set_focus_on_click(False)
        close_button.set_relief(Gtk.ReliefStyle.NONE)
        close_button.connect('clicked', lambda w, p: self._remove_page(p), data.holder)
        close_image = Gtk.Image.new_from_icon_name('stock_close', Gtk.IconSize.INVALID)
        close_image.set_pixel_size(16)
        close_button.add(close_image)

        label_box = Gtk.Box(Gtk.Orientation.HORIZONTAL)
        label_box.pack_start(data.label, False, False, 3)
        label_box.pack_start(close_button, False, False, 0)

        label_eventbox = Gtk.EventBox()
        label_eventbox.add(label_box)
        label_eventbox.show_all()

        data.name = group.entries['name']
        data.background = group.entries['background']

        data.ids = defaultdict(list)
        data.ids[data.name].append(data.name.changed.connect(self._on_name_changed, data))
        data.ids[data.background].append(
            data.background.changed.connect(self._on_background_changed, data))

        self._on_name_changed(data.name, data)
        self._on_background_changed(data.background, data)

        self._page_to_data[data.holder] = data

        if self._widgets.empty.get_parent():
            self._widgets.empty.hide()

        current_idx = self._widgets.notebook.get_current_page()
        current_idx = self._widgets.notebook.insert_page(data.holder, label_eventbox,
                                                         current_idx + 1)

        return current_idx

    def _remove_page(self, page):
        if page == self._widgets.editor.props.parent:
            page.remove(self._widgets.editor)

        self._widgets.notebook.remove_page(self._widgets.notebook.page_num(page))

        del self.monitors.groups[self._page_to_data[page].group]
        del self._page_to_data[page]

        self._update_monitors_list()
        if not self._page_to_data:
            self._widgets.empty.show()

    def _update_monitors_list(self):
        configs = set(group.entries['name'].value for group in self.monitors.groups)
        used_count = 0
        self._widgets.name_combo.get_model().clear()
        for name, item in self._available_monitors:
            used = name in configs
            if used:
                used_count += 1
            item.props.visible = not used
            if not used:
                self._widgets.name_combo.append_text(name)

        show_button = used_count < len(self._available_monitors)
        self._widgets.name_combo.props.button_sensitivity = (Gtk.SensitivityType.ON if show_button
                                                             else Gtk.SensitivityType.OFF)
        self._widgets.editor_add_menu_button.props.visible = show_button
        self._widgets.empty_add_menu_button.props.visible = show_button

    def _on_name_changed(self, entry, data):
        value = entry.value
        markup = None
        error = None

        if not value:
            markup = C_('<i>No name</i>')
            error = C_('The name can\'t be empty. Configuration will not be saved.')
        elif any(data.name != entry and data.name.value == value
                 for data in self._page_to_data.values()):
            error = (C_('"{name}" is already defined. Only last configuration will be saved.')
                     .format(name=value))

        if markup:
            data.label.set_markup(markup)
        else:
            data.label.set_label(value)

        data.name.error = error
        self._update_monitors_list()

    def _on_background_changed(self, entry, data):
        value = entry.value
        if not value or Gdk.RGBA().parse(value):
            entry.error = None
        else:
            entry.error = check_path_accessibility(value)

    def _focus_name_entry(self):
        self._widgets.name.grab_focus()
        self._widgets.name.set_position(0)

    def on_add_button_clicked(self, widget, name=''):
        group = self.monitors.groups.add(name)
        page_idx = self._add_page(group)

        self._widgets.empty.props.visible = not self._page_to_data
        self._widgets.notebook.set_current_page(page_idx)
        if name:
            self._update_monitors_list()

    def on_switch_page(self, notebook, page, page_idx):
        if page == self._widgets.empty:
            buttons = self._widgets.editor_add_menu_button, self._widgets.empty_add_menu_button
        else:
            buttons = self._widgets.empty_add_menu_button, self._widgets.editor_add_menu_button

            old_parent = self._widgets.editor.get_parent()
            if old_parent:
                old_parent.remove(self._widgets.editor)
            page.add(self._widgets.editor)

            data = self._page_to_data[page]
            for key, *__ in self.monitors.EntriesSetup:
                self.monitors.activate(key, data.group.entries[key])

            GLib.idle_add(self._focus_name_entry)

        buttons[0].props.popup = None
        buttons[1].props.popup = self._widgets.available_menu



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2014 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.


from lightdm_gtk_greeter_settings.MultiheadSetupDialog import MultiheadSetupDialog
from lightdm_gtk_greeter_settings import (
    helpers,
    OptionEntry,
    OptionGroup)
from lightdm_gtk_greeter_settings.helpers import (
    WidgetsWrapper)


__all__ = ['MonitorsGroup']


class MonitorsGroup(OptionGroup.BaseGroup):

    GroupPrefix = 'monitor:'
    EntriesSetup = (('name', OptionEntry.StringEntry),
                    ('background', OptionEntry.BackgroundEntry),
                    ('user-background', OptionEntry.BooleanEntry),
                    ('laptop', OptionEntry.BooleanEntry))

    def __init__(self, widgets):
        super().__init__(widgets)
        self._widgets = helpers.WidgetsWrapper(widgets)
        self._groups = []
        self._adapters = {key: OptionGroup.OneToManyEntryAdapter()
                          for key, __ in self.EntriesSetup}
        self._dialog = None

        self._groups_wrapper = helpers.SimpleDictWrapper(
            deleter=self._remove_group,
            add=self._add_group,
            itergetter=lambda: iter(self._groups))

        self._widgets['multihead_label'].connect('activate-link', self._on_label_link_activate)

    def read(self, config):
        for group in self._groups:
            group.clear()
        self._groups.clear()

        for groupname in config:
            if not groupname.startswith(MonitorsGroup.GroupPrefix):
                continue

            monitor = groupname[len(MonitorsGroup.GroupPrefix):].strip()
            self._add_group(monitor, groupname, config)

    def write(self, config, is_changed=None):
        groups = set(group.entries['name'].value for group in self._groups)
        groups_to_remove = tuple(name for name in config
                                 if (name.startswith(self.GroupPrefix) and
                                     name[len(self.GroupPrefix):].strip() not in groups))

        for group in self._groups:
            name = group.entries['name']
            new_name = self.GroupPrefix + ' ' + name.value

            if group.name == new_name:
                def changed_(entry):
                    if entry == name:
                        return False
                    return not is_changed or is_changed(entry)
            else:
                def changed_(entry):
                    return entry != name

            group.name = new_name
            group.write(config, is_changed=changed_)

        for name in groups_to_remove:
            config_group = config[name]
            for key, *__ in self.EntriesSetup:
                del config_group[key]

    def clear(self):
        if not self._groups and not self._adapters:
            return
        for group in self._groups:
            group.clear()
        self._groups.clear()

    def activate(self, key, entry):
        self._adapters[key].activate(entry)

    @property
    def groups(self):
        return self._groups_wrapper

    def _add_group(self, monitor='', groupname='', config=None):
        group = OptionGroup.SimpleGroup(groupname, self._widgets)

        group.entry_added.connect(lambda g, s, e, k: self.entry_added.emit(s, e, k))
        group.entry_removed.connect(lambda g, s, e, k: self.entry_removed.emit(s, e, k))

        group.options = {key: (adapter.new_entry, None)
                         for key, adapter in self._adapters.items()}

        if config:
            group.read(config)

        name = group.entries['name']
        name.enabled = True
        name.value = monitor

        self._groups.append(group)
        return group

    def _remove_group(self, group):
        group.clear()
        self._groups.remove(group)

    def _on_label_link_activate(self, label, uri):
        if not self._dialog:
            self._dialog = MultiheadSetupDialog(self)
            self._dialog.props.transient_for = self._widgets['multihead_label'].get_toplevel()

            for key, klass in self.EntriesSetup:
                self._adapters[key].base_entry = klass(WidgetsWrapper(self._dialog.builder, key))

        self._dialog.run()
        self._dialog.hide()
        return True



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2015 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import configparser
import os
import sys
from collections import OrderedDict
from glob import iglob

from gi.repository import GLib

from lightdm_gtk_greeter_settings import helpers


class Config:

    class ConfigGroup:

        def __init__(self, config):
            self._config = config
            self._items = OrderedDict()

        def __iter__(self):
            return iter(self._items)

        def __contains__(self, item):
            return item in self._items

        def __getitem__(self, item):
            values = self._items.get(item)
            return values[-1][1] if values else None

        def __setitem__(self, item, value):
            if isinstance(value, tuple):
                value, default = value
            else:
                default = None

            values = self._items.get(item)

            if values and values[-1][1] == value:
                return

            if values and values[-1][0] == self._config._output_path:
                if len(values) > 1 and values[-2][1] == value:
                    del values[-1]
                elif default is not None and value == default and len(values) == 1:
                    values.clear()
                else:
                    values[-1] = (self._config._output_path, value)
            elif values is not None:
                if default is None or value != default or (values and values[-1][1] != default):
                    values.append((self._config._output_path, value))
            else:
                if default is None or value != default:
                    self._items[item] = [(self._config._output_path, value)]

        def __delitem__(self, item):
            values = self._items.get(item)
            if values is not None:
                if values and values[-1][0] == self._config._output_path:
                    del values[-1]
                if not values:
                    del self._items[item]

    def __init__(self, base_dir='lightdm', base_name='lightdm-gtk-greeter.conf'):
        self._base_dir = base_dir
        self._base_name = base_name
        self._output_path = helpers.get_config_path()
        self._groups = OrderedDict()
        self._key_values = helpers.SimpleDictWrapper(getter=self._get_key_values)

    def read(self):
        self._groups.clear()

        pathes = []
        pathes += GLib.get_system_data_dirs()
        pathes += GLib.get_system_config_dirs()
        pathes.append(os.path.dirname(os.path.dirname(self._output_path)))

        files = []
        for path in pathes:
            files += sorted(iglob(os.path.join(path, self._base_dir,
                                               self._base_name + '.d', '*.conf')))
            files.append(os.path.join(path, self._base_dir, self._base_name))

        for path in filter(os.path.isfile, files):
            config_file = configparser.RawConfigParser(strict=False, allow_no_value=True)
            try:
                if not config_file.read(path):
                    continue
            except configparser.Error as e:
                print(e, file=sys.stderr)
                continue

            for groupname, values in config_file.items():
                if groupname == 'DEFAULT':
                    continue

                if groupname not in self._groups:
                    self._groups[groupname] = Config.ConfigGroup(self)
                group = self._groups[groupname]

                for key, value in values.items():
                    if value is None:
                        print('[{group}] {key}: Keys without values are not allowed'.format(
                            group=groupname, key=key), file=sys.stderr)
                        continue
                    if key.startswith('-'):
                        key = key[1:]
                        value = None

                    if key in group._items:
                        values = group._items[key]
                        if value is not None or values:
                            values.append((path, value))
                    elif value is not None:
                        group._items[key] = [(path, value)]

    def write(self):
        config_file = configparser.RawConfigParser(strict=False)

        for groupname, group in self._groups.items():
            config_section = None
            for key, values in group._items.items():
                if not values or values[-1][0] != self._output_path:
                    continue

                if values[-1][1] is not None or len(values) > 1:
                    if not config_section:
                        config_file.add_section(groupname)
                        config_section = config_file[groupname]
                    if values[-1][1] is None:
                        config_section['-' + key] = ''
                    else:
                        config_section[key] = values[-1][1]

        with open(self._output_path, 'w') as file:
            config_file.write(file)

    def is_writable(self):
        if os.path.exists(self._output_path) and os.access(self._output_path, os.W_OK):
            return True
        return os.access(os.path.dirname(self._output_path), os.W_OK | os.X_OK)

    def items(self):
        return self._groups.items()

    def allitems(self):
        return ((g, k, items[k]) for (g, items) in self._groups.items() for k in items._items)

    def add_group(self, name):
        if name in self._groups:
            return self._groups[name]
        else:
            return self._groups.setdefault(name, Config.ConfigGroup(self))

    @property
    def key_values(self):
        return self._key_values

    def _get_key_values(self, item):
        group = self._groups.get(item[0])
        if group:
            values = group._items.get(item[1])
            if values is not None:
                return tuple(values)
        return None

    def __iter__(self):
        return iter(self._groups)

    def __getitem__(self, item):
        if isinstance(item, tuple):
            group = self._groups.get(item[0])
            return group[item[1]] if group else None
        return self._groups.get(item)

    def __setitem__(self, item, value):
        if isinstance(item, tuple):
            if not item[0] in self._groups:
                self._groups[item[0]] = Config.ConfigGroup(self)
            self._groups[item[0]][item[1]] = value

    def __delitem__(self, item):
        if isinstance(item, tuple):
            group = self._groups.get(item[0])
            if group is not None:
                del group[item[1]]
            return

        group = self._groups.get(item)
        if group is not None:
            if not group:
                del self._groups[item]
                return

            keys_to_remove = []
            for key, values in group._items.items():
                if values[-1][0] == self._output_path:
                    if len(values) == 1:
                        keys_to_remove.append(key)
                    else:
                        values[-1] = (self._output_path, None)
                elif values:
                    values.append((self._output_path, None))

            if len(keys_to_remove) < len(group._items):
                for key in keys_to_remove:
                    del group._items[key]
            else:
                del self._groups[item]



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2014 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys
from copy import deepcopy
from glob import iglob

from gi.repository import Gtk

from lightdm_gtk_greeter_settings import (
    IconEntry,
    OptionEntry)
from lightdm_gtk_greeter_settings.helpers import (
    C_,
    bool2string,
    string2bool,
    get_data_path,
    get_greeter_version,
    SimpleEnum,
    WidgetsEnum,
    WidgetsWrapper)
from lightdm_gtk_greeter_settings.IndicatorsEntry import (
    EmptyIndicators,
    Indicators,
    LayoutSet,
    Option)


__all__ = ['IndicatorPropertiesDialog']


class IndicatorPath(OptionEntry.StringPathEntry):

    class Row(SimpleEnum):
        Title = ()
        Type = ()
        Icon = ()


class IndicatorIconEntry(IconEntry.IconEntry):

    DefaultValue = ()

    def __init__(self, widgets):
        self._label = widgets['label']
        super().__init__(widgets)

    def _set_value(self, value):
        super()._set_value(self.DefaultValue if value is None else value)
        self._label.set_markup(self._current_item.menuitem.get_label())
        self._image.props.visible = value not in (None, self.DefaultValue)

    def _get_value(self):
        return super()._get_value() or None

    def _get_items(self):
        for item in super()._get_items():
            yield item
        yield -1, (self._update_default, self._ask_default)

    def _update_default(self, value, just_label):
        if just_label or value is not self.DefaultValue:
            return C_('option-entry|indicators', 'Use default value...'), None
        self._image.props.icon_name = ''
        label = C_('option-entry|indicators', '<b>Using default value</b>')
        return label, label

    def _ask_default(self, oldvalue):
        return self.DefaultValue


class IndicatorTypeEntry(OptionEntry.BaseEntry):

    def __init__(self, widgets):
        super().__init__(widgets)

        self._types = widgets['types']
        self._indicator_choice = widgets['indicator_choice']
        self._spacer_choice = widgets['spacer_choice']
        self._separator_choice = widgets['separator_choice']

        self._types.connect('changed', self._emit_changed)
        self._indicator_choice.connect('toggled', self._on_choice_changed, None,
                                       (self._types, widgets['indicator_box']))
        self._spacer_choice.connect('toggled', self._on_choice_changed, Indicators.Spacer)
        self._separator_choice.connect('toggled', self._on_choice_changed, Indicators.Separator)

        self._value = None

    def add_type(self, name, title):
        if name not in EmptyIndicators:
            self._types.append(name, title or name)

    def _get_value(self):
        if self._indicator_choice.props.active:
            return self._types.props.active_id
        else:
            return self._value

    def _set_value(self, value):
        if value == Indicators.Spacer:
            button = self._spacer_choice
        elif value == Indicators.Separator:
            button = self._separator_choice
        else:
            button = self._indicator_choice

        self._value = value
        self._types.set_active_id(value)

        if button.props.active:
            button.toggled()
        else:
            button.props.active = True

    def _on_choice_changed(self, button, value, widgets=[]):
        for w in widgets:
            w.props.sensitive = button.props.active

        if button.props.active:
            self._value = value if value else self._types.props.active_id
            self._emit_changed()


class IndicatorPropertiesDialog(Gtk.Dialog):
    __gtype_name__ = 'IndicatorPropertiesDialog'

    class Widgets(WidgetsEnum):
        add = 'add_button'
        ok = 'ok_button'
        infobar = 'infobar'
        message = 'message'
        common_options = 'common_options_box'
        custom_options = 'custom_options_box'
        path = 'option_path_combo'
        path_model = 'option_path_model'
        hide_disabled = 'option_power_hide_disabled'

    def __new__(cls, *args, **kwargs):
        builder = Gtk.Builder()
        builder.add_from_file(get_data_path('%s.ui' % cls.__name__))
        window = builder.get_object('indicator_properties_dialog')
        window.builder = builder
        builder.connect_signals(window)
        window.init_window(*args, **kwargs)
        return window

    def init_window(self, is_duplicate=None, get_defaults=None, get_name=str):
        self._widgets = self.Widgets(builder=self.builder)
        self._get_defaults = get_defaults
        self._add_indicator = None
        self._is_duplicate = is_duplicate
        self._get_name = get_name
        self._indicator_loaded = False
        self._name = None
        self._reversed = False

        self._name2page = {}
        for i in range(0, self._widgets.custom_options.get_n_pages()):
            page = self._widgets.custom_options.get_nth_page(i)
            name = Gtk.Buildable.get_name(page)
            self._name2page['~' + name.rsplit('_')[-1]] = i

        if get_greeter_version() < 0x020100:
            self._widgets.common_options.props.visible = False

            self._name2page = {
                Indicators.External: self._name2page[Indicators.External],
                Indicators.Text: self._name2page[Indicators.Text]}
            text_prefix = 'option_text_fallback'
        else:
            self._name2page[Indicators.Text] = -1
            text_prefix = 'option_text'

        self._option_type = IndicatorTypeEntry(WidgetsWrapper(self.builder, 'option_type'))
        self._option_text = OptionEntry.StringEntry(WidgetsWrapper(self.builder, text_prefix))
        self._option_image = IndicatorIconEntry(WidgetsWrapper(self.builder, 'option_image'))
        self._option_path = IndicatorPath(WidgetsWrapper(self.builder, 'option_path'))
        self._option_hide_disabled = \
            OptionEntry.BooleanEntry(WidgetsWrapper(self.builder, 'option_hide_disabled'))

        for entry in (self._option_type, self._option_path):
            entry.changed.connect(self._on_option_changed)

        for name in Indicators:
            self._option_type.add_type(name, self._get_name(name))

        # Hiding first column created by Gtk.ComboBoxText
        self._widgets.path.get_cells()[0].props.visible = False

        for path in sorted(iglob(os.path.join(sys.prefix, 'share', 'unity', 'indicators', '*'))):
            name = os.path.basename(path)
            parts = name.rsplit('.', maxsplit=1)
            if len(parts) == 2 and parts[0] == 'com.canonical.indicator':
                name = parts[1]
            row = IndicatorPath.Row._make(Type=IndicatorPath.ItemType.Value,
                                          Title=name,
                                          Icon='application-x-executable')
            self._widgets.path_model.append(row)

        for path in sorted(iglob(os.path.join(sys.prefix, 'lib', 'indicators3', '7', '*.so'))):
            row = IndicatorPath.Row._make(Type=IndicatorPath.ItemType.Value,
                                          Title=os.path.basename(path),
                                          Icon='application-x-executable')
            self._widgets.path_model.append(row)

    def _on_option_changed(self, entry=None):
        if not self._indicator_loaded:
            return

        name = self._option_type.value
        error = None
        warning = None

        if name == Indicators.External:
            if not str(self._option_path.value).strip():
                error = C_('option-entry|indicators', 'Path/Service field is not filled')
        elif name != self._name:
            if self._is_duplicate and self._is_duplicate(name):
                warning = C_('option-entry|indicators',
                             'Indicator "{name}" is already in the list.\n'
                             'It will be overwritten.').format(name=self._get_name(name, name))

        self._widgets.ok.props.sensitive = error is None
        self._widgets.add.props.sensitive = error is None
        self._widgets.infobar.props.visible = error or warning
        self._widgets.message.props.label = error or warning

        if error:
            self._widgets.infobar.props.message_type = Gtk.MessageType.WARNING
        elif warning:
            self._widgets.infobar.props.message_type = Gtk.MessageType.INFO
        else:
            self._widgets.infobar.props.message_type = Gtk.MessageType.OTHER

    def on_option_type_types_changed(self, combo):
        current = self._widgets.custom_options.props.page
        if current != -1:
            self._widgets.custom_options.get_nth_page(current).props.visible = False
        current = self._name2page.get(combo.props.active_id, -1)
        if current != -1:
            self._widgets.custom_options.get_nth_page(current).props.visible = True
            self._widgets.custom_options.props.page = current
        if self._indicator_loaded:
            defaults = self._get_defaults(combo.props.active_id)
            self._option_text.enabled = Option.Text in defaults
            self._option_image.enabled = Option.Image in defaults

    def on_add_clicked(self, widget):
        self._add_callback(self.get_indicator())
        self._options = deepcopy(self._options)
        self._on_option_changed()

    @property
    def add_callback(self):
        return self._add_callback

    @add_callback.setter
    def add_callback(self, value):
        self._add_callback = value
        self._widgets.add.props.visible = value is not None

    def set_indicator(self, options):
        self._indicator_loaded = False
        self._options = deepcopy(options)
        self._name = options[Option.Name]

        self._option_type.value = options[Option.Name]
        self._option_path.value = options.get(Option.Path)

        self._option_text.value = options.get(Option.Text, '')
        self._option_text.enabled = Option.Text in options

        self._option_image.value = options.get(Option.Image)
        self._option_image.enabled = Option.Image in options

        self._reversed = Option.Layout in options and LayoutSet.Reversed in options[Option.Layout]

        hide_disabled = options.get(Option.HideDisabled, bool2string(False))
        self._option_hide_disabled.value = hide_disabled or bool2string(True)

        self._indicator_loaded = True
        self._on_option_changed()

    def get_indicator(self):
        options = self._options

        name = self._option_type.value
        options[Option.Name] = name

        options[Option.Layout] = set()
        if name not in EmptyIndicators:
            if self._option_text.enabled:
                options[Option.Text] = self._option_text.value or None
                options[Option.Layout].add(LayoutSet.Text)
            if self._option_image.enabled:
                options[Option.Image] = self._option_image.value or None
                options[Option.Layout].add(LayoutSet.Image)
            if self._option_text.enabled and self._option_image.enabled and self._reversed:
                options[Option.Layout].add(LayoutSet.Reversed)

        if LayoutSet.Text not in options[Option.Layout] and Option.Text in options:
            del options[Option.Text]
        if LayoutSet.Image not in options[Option.Layout] and Option.Image in options:
            del options[Option.Image]

        if name == Indicators.External:
            options[Option.Path] = self._option_path.value
        else:
            options.pop(Option.Path, None)

        if name == Indicators.Power and string2bool(self._option_hide_disabled.value):
            options[Option.HideDisabled] = None
        elif Option.HideDisabled in options:
            options.pop(Option.HideDisabled, None)

        return options



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2014 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import (
    GObject,
    Gtk)

from lightdm_gtk_greeter_settings.helpers import (
    get_data_path,
    NC_,
    SimpleEnum,
    WidgetsEnum)


__all__ = ['IconChooserDialog']


class ContextRow(SimpleEnum):
    Name = ()
    Standard = ()
    Title = ()


class IconRow(SimpleEnum):
    Name = ()
    Standard = ()
    Context = ()


class IconChooserDialog(Gtk.Dialog):

    __gtype_name__ = 'IconChooserDialog'

    def __new__(cls):
        builder = Gtk.Builder()
        builder.add_from_file(get_data_path('%s.ui' % cls.__name__))
        window = builder.get_object('icon_chooser_dialog')
        window.builder = builder
        builder.connect_signals(window)
        window.init_window()
        return window

    class Widgets(WidgetsEnum):
        name = 'name_entry'
        preview = 'preview_image'
        standard = 'standard_toggle'
        ok = 'ok_button'
        cancel = 'cancel_button'
        contexts_view = 'contexts_view'
        contexts_selection = 'contexts_selection'
        contexts_model = 'contexts_model'
        contexts_filter = 'contexts_filter'
        icons_view = 'icons_view'
        icons_selection = 'icons_selection'
        icons_model = 'icons_model'
        icons_sorted = 'icons_sorted'
        icons_filter = 'icons_filter'

    builder = None

    def init_window(self):
        self._widgets = self.Widgets(builder=self.builder)
        self._widgets_to_disable = (self._widgets.name, self._widgets.standard,
                                    self._widgets.contexts_view, self._widgets.icons_view,
                                    self._widgets.ok, self._widgets.cancel)

        self._icons_loaded = False
        self._icon_to_select = None
        self._icon_filter_standard = True
        self._icon_filter_context = None

        self._widgets.contexts_view.set_row_separator_func(self._contexts_separator_callback, None)
        self._widgets.contexts_filter.set_visible_func(self._contexts_visible_callback)

        self._widgets.icons_filter.set_visible_func(self._icons_filter_visible_callback)
        self._widgets.icons_sorted.set_sort_column_id(0, Gtk.SortType.ASCENDING)

        self._reload()

    def _reload(self):
        for w in self._widgets_to_disable:
            w.props.sensitive = False
        GObject.idle_add(self._read_icons)

    def _read_icons(self):
        theme = Gtk.IconTheme.get_default()
        standard_contexts = set(name for name, title in StandardContexts)

        self._widgets.contexts_model.clear()
        for name, title in StandardContexts:
            row = ContextRow._make(Name=name, Standard=True, Title=title)
            self._widgets.contexts_model.append(row)

        for name in theme.list_contexts():
            if name not in standard_contexts:
                row = ContextRow._make(Name=name, Standard=False, Title=name)
                self._widgets.contexts_model.append(row)

        self._widgets.icons_model.clear()
        for context in theme.list_contexts():
            for icon in theme.list_icons(context):
                row = IconRow._make(Name=icon, Standard=icon in StandardIconNames, Context=context)
                self._widgets.icons_model.append(row)

        self._icons_loaded = True
        if self._icon_to_select:
            self.select_icon(self._icon_to_select)
            self._icon_to_select = None

        for w in self._widgets_to_disable:
            w.props.sensitive = True

        return False

    def _update_contexts_filter(self):
        self._widgets.contexts_filter.refilter()
        self._update_icons_filter()

    def _update_icons_filter(self):
        self._widgets.icons_view.props.model = None
        model, rowiter = self._widgets.contexts_selection.get_selected()
        if rowiter:
            self._icon_filter_standard = self._widgets.standard.props.active
            self._icon_filter_context = model[rowiter][ContextRow.Name]
            self._widgets.icons_filter.refilter()
            self._widgets.icons_view.props.model = self._widgets.icons_sorted

    def _contexts_visible_callback(self, model, rowiter, data):
        if not self._widgets.standard.props.active:
            return True
        return model[rowiter][ContextRow.Standard]

    def _contexts_separator_callback(self, model, rowiter, data):
        return not model[rowiter][ContextRow.Name] and not model[rowiter][ContextRow.Title]

    def _icons_filter_visible_callback(self, model, rowiter, data):
        if self._icon_filter_standard and not model[rowiter][IconRow.Standard]:
            return False
        if not self._icon_filter_context:
            return True
        return model[rowiter][IconRow.Context] == self._icon_filter_context

    def get_selected_icon(self):
        return self._widgets.name.props.text

    def select_icon(self, name):
        if not self._icons_loaded:
            self._icon_to_select = name
            return

        if name not in StandardIconNames:
            self._widgets.standard.props.active = False
        self._widgets.contexts_selection.select_path(0)

        for row in self._widgets.icons_sorted:
            if row[IconRow.Name] == name:
                self._widgets.icons_view.set_cursor(row.path)
                self._widgets.icons_selection.select_path(row.path)
                break
        else:
            self._widgets.name.props.text = name

    def on_icons_selection_changed(self, selection):
        model, rowiter = selection.get_selected()
        if rowiter:
            self._widgets.name.props.text = model[rowiter][IconRow.Name]

    def on_contexts_selection_changed(self, selection):
        self._widgets.icons_selection.unselect_all()
        self._update_icons_filter()

    def on_standard_toggled(self, toggle):
        self._update_contexts_filter()

    def on_name_changed(self, entry):
        name = entry.props.text
        if not Gtk.IconTheme.get_default().has_icon(name):
            name = ''
        self._widgets.preview.props.icon_name = name


# http://standards.freedesktop.org/icon-naming-spec/icon-naming-spec-latest.html

StandardContexts = (
    (None, NC_('icon-dialog', 'All contexts')),
    (None, ''),  # separator
    ('Actions', NC_('icon-dialog', 'Actions')),
    ('Applications', NC_('icon-dialog', 'Applications')),
    ('Categories', NC_('icon-dialog', 'Categories')),
    ('Devices', NC_('icon-dialog', 'Devices')),
    ('Emblems', NC_('icon-dialog', 'Emblems')),
    ('Emotes', NC_('icon-dialog', 'Emoticons')),
    ('International', NC_('icon-dialog', 'International')),
    ('MimeTypes', NC_('icon-dialog', 'MIME Types')),
    ('Places', NC_('icon-dialog', 'Places')),
    ('Status', NC_('icon-dialog', 'Status')))

StandardIconNames = {
    # Actions
    'address-book-new', 'application-exit', 'appointment-new',
    'call-start', 'call-stop', 'contact-new', 'document-new',
    'document-open', 'document-open-recent', 'document-page-setup',
    'document-print', 'document-print-preview', 'document-properties',
    'document-revert', 'document-save', 'document-save-as',
    'document-send', 'edit-clear', 'edit-copy', 'edit-cut', 'edit-delete',
    'edit-find', 'edit-find-replace', 'edit-paste', 'edit-redo',
    'edit-select-all', 'edit-undo', 'folder-new', 'format-indent-less',
    'format-indent-more', 'format-justify-center', 'format-justify-fill',
    'format-justify-left', 'format-justify-right',
    'format-text-direction-ltr', 'format-text-direction-rtl',
    'format-text-bold', 'format-text-italic', 'format-text-underline',
    'format-text-strikethrough', 'go-bottom', 'go-down', 'go-first',
    'go-home', 'go-jump', 'go-last', 'go-next', 'go-previous', 'go-top',
    'go-up', 'help-about', 'help-contents', 'help-faq', 'insert-image',
    'insert-link', 'insert-object', 'insert-text', 'list-add',
    'list-remove', 'mail-forward', 'mail-mark-important', 'mail-mark-junk',
    'mail-mark-notjunk', 'mail-mark-read', 'mail-mark-unread',
    'mail-message-new', 'mail-reply-all', 'mail-reply-sender', 'mail-send',
    'mail-send-receive', 'media-eject', 'media-playback-pause',
    'media-playback-start', 'media-playback-stop', 'media-record',
    'media-seek-backward', 'media-seek-forward', 'media-skip-backward',
    'media-skip-forward', 'object-flip-horizontal', 'object-flip-vertical',
    'object-rotate-left', 'object-rotate-right', 'process-stop',
    'system-lock-screen', 'system-log-out', 'system-run', 'system-search',
    'system-reboot', 'system-shutdown', 'tools-check-spelling',
    'view-fullscreen', 'view-refresh', 'view-restore',
    'view-sort-ascending', 'view-sort-descending', 'window-close',
    'window-new', 'zoom-fit-best', 'zoom-in', 'zoom-original', 'zoom-out',
    # StandardApplicationIcons
    'accessories-calculator', 'accessories-character-map',
    'accessories-dictionary', 'accessories-text-editor', 'help-browser',
    'multimedia-volume-control', 'preferences-desktop-accessibility',
    'preferences-desktop-font', 'preferences-desktop-keyboard',
    'preferences-desktop-locale', 'preferences-desktop-multimedia',
    'preferences-desktop-screensaver', 'preferences-desktop-theme',
    'preferences-desktop-wallpaper', 'system-file-manager',
    'system-software-install', 'system-software-update',
    'utilities-system-monitor', 'utilities-terminal',
    # StandardCategoryIcons
    'applications-accessories', 'applications-development',
    'applications-engineering', 'applications-games',
    'applications-graphics', 'applications-internet',
    'applications-multimedia', 'applications-office', 'applications-other',
    'applications-science', 'applications-system',
    'applications-utilities', 'preferences-desktop',
    'preferences-desktop-peripherals', 'preferences-desktop-personal',
    'preferences-other', 'preferences-system',
    'preferences-system-network', 'system-help',
    # StandardDeviceIcons
    'audio-card', 'audio-input-microphone', 'battery', 'camera-photo',
    'camera-video', 'camera-web', 'computer', 'drive-harddisk',
    'drive-optical', 'drive-removable-media', 'input-gaming',
    'input-keyboard', 'input-mouse', 'input-tablet', 'media-flash',
    'media-floppy', 'media-optical', 'media-tape', 'modem',
    'multimedia-player', 'network-wired', 'network-wireless', 'pda',
    'phone', 'printer', 'scanner', 'video-display',
    # StandardEmblemIcons
    'emblem-default', 'emblem-documents', 'emblem-downloads',
    'emblem-favorite', 'emblem-important', 'emblem-mail', 'emblem-photos',
    'emblem-readonly', 'emblem-shared', 'emblem-symbolic-link',
    'emblem-synchronized', 'emblem-system', 'emblem-unreadable',
    # StandardEmotionIcons
    'face-angel', 'face-angry', 'face-cool', 'face-crying',
    'face-devilish', 'face-embarrassed', 'face-kiss', 'face-laugh',
    'face-monkey', 'face-plain', 'face-raspberry', 'face-sad', 'face-sick',
    'face-smile', 'face-smile-big', 'face-smirk', 'face-surprise',
    'face-tired', 'face-uncertain', 'face-wink', 'face-worried',
    # StandardInternationalIcons
    'flag-aa',
    # StandardMIMETypeIcons
    'application-x-executable', 'audio-x-generic', 'font-x-generic',
    'image-x-generic', 'package-x-generic', 'text-html', 'text-x-generic',
    'text-x-generic-template', 'text-x-script', 'video-x-generic',
    'x-office-address-book', 'x-office-calendar', 'x-office-document',
    'x-office-presentation', 'x-office-spreadsheet',
    # StandardPlaceIcons
    'folder', 'folder-remote', 'network-server', 'network-workgroup',
    'start-here', 'user-bookmarks', 'user-desktop', 'user-home',
    'user-trash',
    # StandardStatusIcons
    'appointment-missed', 'appointment-soon', 'audio-volume-high',
    'audio-volume-low', 'audio-volume-medium', 'audio-volume-muted',
    'battery-caution', 'battery-low', 'dialog-error', 'dialog-information',
    'dialog-password', 'dialog-question', 'dialog-warning',
    'folder-drag-accept', 'folder-open', 'folder-visiting',
    'image-loading', 'image-missing', 'mail-attachment', 'mail-unread',
    'mail-read', 'mail-replied', 'mail-signed', 'mail-signed-verified',
    'media-playlist-repeat', 'media-playlist-shuffle', 'network-error',
    'network-idle', 'network-offline', 'network-receive',
    'network-transmit', 'network-transmit-receive', 'printer-error',
    'printer-printing', 'security-high', 'security-medium', 'security-low',
    'software-update-available', 'software-update-urgent', 'sync-error',
    'sync-synchronizing', 'task-due', 'task-past-due', 'user-available',
    'user-away', 'user-idle', 'user-offline', 'user-trash-full',
    'weather-clear', 'weather-clear-night', 'weather-few-clouds',
    'weather-few-clouds-night', 'weather-fog', 'weather-overcast',
    'weather-severe-alert', 'weather-showers', 'weather-showers-scattered',
    'weather-snow', 'weather-storm'}



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2014 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.


import collections
import os
import shlex
import sys
from functools import partialmethod
from glob import iglob
from itertools import chain
from locale import gettext as _

from gi.repository import (
    Gdk,
    GLib,
    Gtk)
from gi.repository import Pango
from gi.repository.GObject import markup_escape_text as escape_markup

from lightdm_gtk_greeter_settings import (
    Config,
    helpers,
    IconEntry,
    IndicatorsEntry,
    OptionEntry,
    PositionEntry)
from lightdm_gtk_greeter_settings.helpers import (
    C_,
    string2bool,
    SimpleEnum,
    WidgetsEnum,
    WidgetsWrapper)
from lightdm_gtk_greeter_settings.MonitorsGroup import MonitorsGroup
from lightdm_gtk_greeter_settings.OptionGroup import SimpleGroup


__all__ = ['GtkGreeterSettingsWindow',
           'WindowMode']


class WindowMode(SimpleEnum):
    Default = 'default'
    Embedded = 'embedded'
    GtkHeader = 'gtk-header'


InitialValue = collections.namedtuple('InitialValue', ('value', 'enabled'))


class GtkGreeterSettingsWindow(Gtk.Window):

    __gtype_name__ = 'GtkGreeterSettingsWindow'

    class Widgets(WidgetsEnum):
        apply = 'apply_button'
        reload = 'reset_button'
        close = 'close_button'
        buttons = 'dialog_buttons'
        content = 'content_box'
        infobar = 'infobar'
        infobar_label = 'infobar_label'
        multihead_label = 'multihead_label'

    def __new__(cls, mode=WindowMode.Default):
        builder = Gtk.Builder()
        builder.add_from_file(helpers.get_data_path('%s.ui' % cls.__name__))
        window = builder.get_object('settings_window')
        window.builder = builder
        window.mode = mode
        builder.connect_signals(window)
        window.init_window()
        return window

    builder = None
    mode = WindowMode.Default

    GreeterGroupSetup = {
        # Appearance
        'theme-name': (OptionEntry.StringEntry, ''),
        'icon-theme-name': (OptionEntry.StringEntry, ''),
        'font-name': (OptionEntry.FontEntry, 'Sans 10'),
        'xft-antialias': (OptionEntry.BooleanEntry, None),
        'xft-dpi': (OptionEntry.StringEntry, None),
        'xft-rgba': (OptionEntry.ChoiceEntry, None),
        'xft-hintstyle': (OptionEntry.ChoiceEntry, None),
        'background': (OptionEntry.BackgroundEntry, '#000000'),
        'user-background': (OptionEntry.BooleanEntry, 'true'),
        'hide-user-image': (OptionEntry.InvertedBooleanEntry, 'false'),
        'default-user-image': (IconEntry.IconEntry, '#avatar-default'),
        # Panel
        'clock-format': (OptionEntry.ClockFormatEntry, '%a, %H:%M'),
        'indicators': (IndicatorsEntry.IndicatorsEntry,
                       '~host;~spacer;~clock;~spacer;~language;~session;~a11y;~power'),
        # Position
        'position': (PositionEntry.PositionEntry, '50%,center'),
        # Misc
        'screensaver-timeout': (OptionEntry.AdjustmentEntry, '60'),
        'keyboard': (OptionEntry.StringPathEntry, ''),
        'reader': (OptionEntry.StringPathEntry, ''),
        'a11y-states': (OptionEntry.AccessibilityStatesEntry, ''),
        'allow-debugging': (OptionEntry.BooleanEntry, 'false'), }

    entries_setup = {
        ('greeter', 'allow-debugging'): ('changed',),
        ('greeter', 'background'): ('changed',),
        ('greeter', 'default-user-image'): ('changed',),
        ('greeter', 'screensaver-timeout'): ('setup', 'get', 'set'),
        ('greeter', 'theme-name'): ('setup', 'changed'),
        ('greeter', 'icon-theme-name'): ('setup', 'changed'),
        ('greeter', 'keyboard'): ('changed',),
        ('greeter', 'reader'): ('changed',)}

    def init_window(self):
        self._widgets = self.Widgets(builder=self.builder)

        if self.mode == WindowMode.Embedded:
            self.on_entry_removed = self.on_entry_removed_embedded
            self.on_entry_changed = self.on_entry_changed_embedded
            self._write = self._write_embedded

            self._widgets.buttons.hide()
            self._widgets.content.reorder_child(self._widgets.infobar, 0)
            self._widgets.content.connect('destroy', self.on_destroy, True)
            # Socket/Plug focus issues workaround
            self._widgets.multihead_label.connect('button-press-event', self.on_multihead_click)
        elif self.mode == WindowMode.GtkHeader:
            for button in (self._widgets.apply, self._widgets.reload):
                self._widgets.buttons.remove(button)
                button.set_label('')
                button.set_always_show_image(True)
            self._widgets.buttons.hide()

            header = Gtk.HeaderBar()
            header.set_show_close_button(True)
            header.props.title = self.get_title()
            header.pack_start(self._widgets.reload)
            header.pack_start(self._widgets.apply)
            header.show_all()

            self.set_titlebar(header)

        self._config = Config.Config()

        self._entry_menu = None
        self._initial_values = {}
        self._entries = None

        self._changed_entries = None
        self._new_entries = None
        self._removed_entries = None

        self._groups = (
            SimpleGroup('greeter', WidgetsWrapper(self.builder, 'greeter'), self.GreeterGroupSetup),
            MonitorsGroup(self.builder))

        for group in self._groups:
            group.entry_added.connect(self.on_entry_added)
            group.entry_removed.connect(self.on_entry_removed)

        self._allow_edit = self._config.is_writable()
        self._update_apply_button()

        if not self._allow_edit:
            self._set_message(_('You don\'t have permissions to change greeter configuration'),
                              Gtk.MessageType.WARNING)
            if self.mode != WindowMode.Embedded:
                helpers.show_message(
                    text=_('No permissions to save configuration'),
                    secondary_text=_(
                        'It seems that you don\'t have permissions to write to '
                        'file:\n{path}\n\nTry to run this program using "sudo" '
                        'or "pkexec"').format(path=helpers.get_config_path()),
                    message_type=Gtk.MessageType.WARNING)

        self._read()

    def _set_message(self, message, type_=Gtk.MessageType.INFO):
        if not message:
            self._widgets.infobar.hide()

        self._widgets.infobar.props.message_type = type_
        self._widgets.infobar_label.props.label = message
        self._widgets.infobar.show()

    def _update_apply_button(self):
        allow = (self._allow_edit and
                 (self._changed_entries or self._new_entries or self._removed_entries))
        self._widgets.apply.props.sensitive = allow

    def _read(self):
        self._config.read()
        self._changed_entries = None
        self._new_entries = None
        self._removed_entries = None

        for group in self._groups:
            group.read(self._config)

        self._initial_values = {entry: InitialValue(entry.value, entry.enabled)
                                for entry in self._initial_values.keys()}

        self._changed_entries = set()
        self._new_entries = set()
        self._removed_entries = set()

        self._update_apply_button()

    def _write(self):
        changed = self._changed_entries | self._new_entries
        for group in self._groups:
            group.write(self._config, changed.__contains__)

        if self.mode != WindowMode.Embedded:
            for entry in changed:
                self._initial_values[entry] = InitialValue(entry.value, entry.enabled)

        self._changed_entries.clear()
        self._new_entries.clear()
        self._removed_entries.clear()

        try:
            self._config.write()
        except OSError as e:
            helpers.show_message(e, Gtk.MessageType.ERROR)

        self._update_apply_button()

    _write_timeout_id = None

    def _write_embedded(self, delay=750):
        if self._write_timeout_id:
            GLib.Source.remove(self._write_timeout_id)
        if delay:
            self._write_timeout_id = GLib.timeout_add(delay, self.on_write_timeout)
        else:
            self._write_timeout_id = None
            self.on_write_timeout()

    def on_write_timeout(self):
        self._write_timeout_id = None
        self.__class__._write(self)
        return False

    def on_entry_added(self, group, source, entry, key):
        if isinstance(source, SimpleGroup) and (source.name, key) in self.entries_setup:
            for action in self.entries_setup[(source.name, key)]:
                fname = 'on_entry_%s_%s_%s' % (action, source.name, key)
                f = getattr(self, fname.replace('-', '_'))
                if action == 'setup':
                    f(entry)
                else:
                    entry.connect(action, f)

        entry.show_menu.connect(self.on_show_menu, source, key)
        entry.changed.connect(self.on_entry_changed)

        self._initial_values[entry] = InitialValue(entry.value, entry.enabled)
        self.on_entry_changed(entry, forced=True)

        if self._new_entries is not None:
            self._new_entries.add(entry)

    def on_entry_removed(self, source, group, entry, key):
        if self._changed_entries is None:
            return

        self._initial_values.pop(entry, None)
        if entry in self._new_entries:
            self._new_entries.discard(entry)
            self._changed_entries.discard(entry)
        else:
            self._removed_entries.add(entry)

        self._update_apply_button()

    def on_entry_removed_embedded(self, source, group, entry, key):
        if self._removed_entries is None:
            return

        self._initial_values.pop(entry, None)
        self._removed_entries.add(entry)
        if self._allow_edit:
            self._write()

    def on_entry_changed(self, entry, forced=False):
        if self._changed_entries is None:
            return

        initial = self._initial_values[entry]
        if forced or entry.enabled != initial.enabled or \
           (entry.enabled and entry.value != initial.value):
            self._changed_entries.add(entry)
        else:
            self._changed_entries.discard(entry)

        self._update_apply_button()

    def on_entry_changed_embedded(self, entry, forced=False):
        if self._changed_entries is not None:
            self._changed_entries.add(entry)
            if self._allow_edit:
                self._write()

    def on_entry_reset_clicked(self, item):
        entry, value, enabled = item._reset_entry_data
        if enabled is None:
            entry.value = value
        else:
            entry.enabled = enabled

    def on_entry_fix_clicked(self, item):
        entry, action = item._fix_entry_data
        action(entry)

    def on_show_menu(self, entry, group, key):

        def new_item(activate=None, width=90):
            item = Gtk.MenuItem('')
            label = item.get_child()
            label.props.use_markup = True
            label.props.ellipsize = Pango.EllipsizeMode.END
            label.props.max_width_chars = width
            if activate:
                item.connect('activate', activate)
            else:
                item.props.sensitive = False
            return item

        if not self._entry_menu:
            class EntryMenu:
                menu = Gtk.Menu()
                group = new_item()
                value = new_item()
                file = new_item()
                error_separator = Gtk.SeparatorMenuItem()
                error = new_item()
                error_action = new_item(self.on_entry_fix_clicked)
                reset_separator = Gtk.SeparatorMenuItem()
                initial = new_item(self.on_entry_reset_clicked)
                default = new_item(self.on_entry_reset_clicked)
                other = []

                menu.append(group)
                menu.append(value)
                menu.append(file)
                menu.append(error_separator)
                menu.append(error)
                menu.append(error_action)
                menu.append(reset_separator)
                menu.append(initial)
                menu.append(default)
                menu.show_all()

            self._entry_menu = EntryMenu()

        def format_value(value=None, enabled=True):
            if not enabled:
                return _('<i>disabled</i>')
            if value == '':
                return _('<i>empty string</i>')
            elif value is None:
                return _('<i>None</i>')
            else:
                return escape_markup(str(value))

        menu = self._entry_menu

        # [group]
        if group.name:
            menu.group.props.label = '[{group}]'.format(group=group.name)
            menu.group.show()
        else:
            menu.group.hide()

        # key = value
        if entry.enabled:
            menu.value.props.label = '{key} = {value}'.format(
                key=key, value=format_value(value=entry.value, enabled=entry.enabled))
        else:
            menu.value.props.label = '# {key} ='.format(key=key)

        # File with key definition
        config_values = self._config.key_values[group.name, key]
        if entry not in self._changed_entries and \
           config_values and config_values[-1][0] != helpers.get_config_path():
            menu.file.props.label = _('Value defined in file: {path}')\
                .format(path=escape_markup(config_values[-1][0]))
            menu.file.set_tooltip_text(config_values[-1][0])
            menu.file.show()
        else:
            menu.file.hide()

        # Error message
        error = entry.error
        error_action = None
        if error:
            aname = ('get_entry_fix_%s_%s' % (group.name, key)).replace('-', '_')
            get_fix = getattr(self, aname, None)
            if get_fix:
                label, error_action = get_fix(entry)
                if label:
                    menu.error_action.props.label = label or ''
                if error_action:
                    menu.error_action._fix_entry_data = entry, error_action
            menu.error.set_label(escape_markup(error))

        menu.error.props.visible = error is not None
        menu.error_action.props.visible = error_action is not None
        menu.error_separator.props.visible = error_action is not None

        # Reset to initial value
        initial = self._initial_values[entry]
        if initial.enabled != entry.enabled or initial.value != entry.value:
            if entry.enabled != initial.enabled and not initial.enabled:
                menu.initial._reset_entry_data = entry, None, initial.enabled
            else:
                menu.initial._reset_entry_data = entry, initial.value, None

            value = format_value(value=initial.value, enabled=initial.enabled)
            menu.initial.set_tooltip_markup(value)
            menu.initial.props.visible = True
            menu.initial.props.label = \
                _('Reset to initial value: <b>{value}</b>').format(value=value)
        else:
            menu.initial.props.visible = False

        # Reset to default value
        default = group.defaults[key]
        if default is not None and entry.value != default:
            value = format_value(value=default)
            menu.default._reset_entry_data = entry, default, None
            menu.default.set_tooltip_markup(value)
            menu.default.props.visible = True
            menu.default.props.label = \
                _('Reset to default value: <b>{value}</b>').format(value=value)
        else:
            menu.default.props.visible = False

        # Reset to values from all other (.conf
        item_idx = 0
        if config_values and len(config_values) > 1:
            values = {None, default, self._initial_values[entry].value, entry.value}
            for __, value in config_values[:-1]:
                if value in values:
                    continue

                if len(menu.other) <= item_idx:
                    item = new_item(self.on_entry_reset_clicked)
                    menu.other.append(item)
                    menu.menu.append(item)
                else:
                    item = menu.other[item_idx]
                item._reset_entry_data = entry, value, None
                value = format_value(value=value)
                item.set_tooltip_markup(value)
                item.props.label = _('Reset to value: <b>{value}</b>').format(value=value)
                item.show()
                item_idx += 1
        for item in menu.other[item_idx:]:
            item.hide()

        menu.reset_separator.props.visible = \
            menu.initial.props.visible or menu.default.props.visible or \
            any(item.props.visible for item in menu.other)

        self._entry_menu.menu.popup(None, None, None, None, 0, Gtk.get_current_event_time())

    # [greeter] screensaver-timeout
    def on_entry_setup_greeter_screensaver_timeout(self, entry):
        view = entry.widgets['view']
        adjustment = entry.widgets['adjustment']
        end_label = entry.widgets['end-label']
        for mark in chain((range(10, 61, 10)),
                          (range(69, int(adjustment.props.upper), 10))):
            view.add_mark(mark, Gtk.PositionType.BOTTOM, None)
        total = int(adjustment.props.upper - 60) + 1
        end_label.props.label = C_('option|greeter|screensaver-timeout',
                                   '{count} min').format(count=total)

        view.connect('format-value',
                     self.on_entry_format_scale_greeter_screensaver_timeout,
                     adjustment)

    def on_entry_get_greeter_screensaver_timeout(self, entry=None, value=None):
        try:
            value = int(float(value))
        except ValueError:
            value = 60

        if value > 60:
            return (value - 59) * 60
        return value

    def on_entry_set_greeter_screensaver_timeout(self, entry=None, value=None):
        try:
            value = int(float(value))
        except ValueError:
            value = 60

        if value > 60:
            return value // 60 + 59
        return value

    def on_entry_format_scale_greeter_screensaver_timeout(self, scale, value, adjustment):
        if value != adjustment.props.lower and value != adjustment.props.upper:
            value = self.on_entry_get_greeter_screensaver_timeout(value=value)
            return '%02d:%02d' % (value // 60, value % 60)
        return ''

    # [greeter] theme-name
    # LP: #1709864, Support gtk-3.* themes
    GtkThemesPattern = (sys.prefix, 'share', 'themes', '*', 'gtk-3.*', 'gtk.css')

    def on_entry_setup_greeter_theme_name(self, entry, pattern=GtkThemesPattern):
        values = entry.widgets['values']
        themes = []
        idx = pattern.index('*') - len(pattern)
        for path in sorted(iglob(os.path.join(*pattern))):
            theme = path.split(os.path.sep)[idx]
            if theme not in themes:
                themes.append(theme)

        themes = sorted(themes, key=lambda theme: theme.lower())

        for theme in themes:
            values.append_text(theme)

    def on_entry_changed_greeter_theme_name(self, entry, pattern=GtkThemesPattern):
        value = entry.value
        if value:
            path = (p if p != '*' else value.strip() for p in pattern)
            entry.error = helpers.check_path_accessibility(os.path.join(*path))
        else:
            entry.error = None

    # [greeter] icon-theme-name
    IconThemesPattern = (sys.prefix, 'share', 'icons', '*', 'index.theme')
    on_entry_setup_greeter_icon_theme_name = partialmethod(on_entry_setup_greeter_theme_name,
                                                           pattern=IconThemesPattern)

    on_entry_changed_greeter_icon_theme_name = partialmethod(on_entry_changed_greeter_theme_name,
                                                             pattern=IconThemesPattern)

    # [greeter] allow-debugging
    def on_entry_changed_greeter_allow_debugging(self, entry):
        gtk_version = Gtk.MAJOR_VERSION, Gtk.MINOR_VERSION, Gtk.MICRO_VERSION
        minimal_version = 3, 14, 0
        if gtk_version < minimal_version and string2bool(entry.value):
            entry.error = C_('option|greeter|allow-debugging',
                             'GtkInspector is not available on your system\n'
                             'Gtk version: {current} < {minimal}').format(
                current=gtk_version, minimal=minimal_version)
        else:
            entry.error = None

    # [greeter] default-user-image
    def on_entry_changed_greeter_default_user_image(self, entry):
        value = entry.value
        if value.startswith('#'):
            entry.error = None
        else:
            entry.error = helpers.check_path_accessibility(value)

    def get_entry_fix_greeter_default_user_image(self, entry):
        value = entry.value
        if not entry.error or value.startswith('#'):
            return None, None
        return None, None

    # [greeter] background
    def on_entry_changed_greeter_background(self, entry):
        value = entry.value
        if not value or Gdk.RGBA().parse(value):
            entry.error = None
        else:
            entry.error = helpers.check_path_accessibility(value)

    # [greeter] keyboard
    def on_entry_changed_greeter_keyboard(self, entry):
        error = None
        if entry.enabled:
            value = entry.value
            if os.path.isabs(value):
                argv = shlex.split(value)
                error = helpers.check_path_accessibility(argv[0], executable=True)
        entry.error = error

    # [greeter] reader
    on_entry_changed_greeter_reader = on_entry_changed_greeter_keyboard

    def on_multihead_click(self, label, event):
        if event.button == 1:
            label.emit('activate-link', '')
            return True
        return False

    def on_destroy(self, widget, write=False):
        if write and self._write_timeout_id:
            self._write_embedded(delay=None)
        Gtk.main_quit()

    def on_apply_clicked(self, *unused):
        self._write()

    def on_reset_clicked(self, *unused):
        self._read()

    def on_close_clicked(self, *unused):
        self.destroy()



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2014 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.


import configparser
import glob
import locale
import os
import pwd
import stat

from collections import (
    namedtuple,
    OrderedDict,
    defaultdict)
from itertools import (
    chain,
    accumulate)
from locale import gettext as _

from gi.repository import (
    GdkPixbuf,
    GLib,
    GObject,
    Gtk,
    Pango)


__license__ = 'GPL-3'
__version__ = 'dev'
__data_directory__ = '../data/'
__config_path__ = 'lightdm/lightdm-gtk-greeter.conf'


try:
    from . installation_config import (
        __version__,
        __data_directory__,
        __config_path__)
except ImportError:
    pass


__all__ = [
    'bool2string',
    'C_',
    'clamp',
    'check_path_accessibility',
    'DefaultValueDict',
    'file_is_readable_by_greeter',
    'get_config_path',
    'get_data_path',
    'get_greeter_version'
    'get_markup_error',
    'get_version',
    'ModelRowEnum',
    'NC_',
    'pixbuf_from_file_scaled_down',
    'set_image_from_path',
    'show_message',
    'SimpleEnum',
    'SimpleDictWrapper',
    'string2bool',
    'TreeStoreDataWrapper',
    'WidgetsEnum',
    'WidgetsWrapper']


def C_(context, message):
    separator = '\x04'
    message_with_context = '{}{}{}'.format(context, separator, message)
    result = locale.gettext(message_with_context)
    if separator in result:
        return message
    return result


def NC_(context, message):
    return message


def get_data_path(*parts):
    return os.path.abspath(os.path.join(os.path.dirname(__file__),
                                        __data_directory__, *parts))


def get_config_path():
    return os.path.abspath(__config_path__)


def get_version():
    return __version__


def get_greeter_version():
    try:
        return get_greeter_version._version
    except AttributeError:
        try:
            get_greeter_version._version = int(os.getenv('GTK_GREETER_VERSION', '0x010900'), 16)
        except ValueError:
            get_greeter_version._version = 0x010900

    return get_greeter_version._version


def bool2string(value, skip_none=False):
    if isinstance(value, str):
        value = string2bool(value)
    return 'true' if value else 'false' if not skip_none or value is not None else None


def string2bool(value, fallback=False):
    if isinstance(value, str):
        if value in ('true', 'yes', '1'):
            return True
        if value in ('false', 'no', '0'):
            return False
    return fallback


def show_message(**kwargs):
    dialog = Gtk.MessageDialog(parent=Gtk.Window.list_toplevels()[0],
                               buttons=Gtk.ButtonsType.CLOSE, **kwargs)
    dialog.run()
    dialog.destroy()


def pixbuf_from_file_scaled_down(path, width, height):
    pixbuf = GdkPixbuf.Pixbuf.new_from_file(path)
    scale = max(pixbuf.props.width / width, pixbuf.props.height / height)

    if scale > 1:
        return GdkPixbuf.Pixbuf.scale_simple(pixbuf,
                                             pixbuf.props.width / scale,
                                             pixbuf.props.height / scale,
                                             GdkPixbuf.InterpType.BILINEAR)
    return pixbuf


def set_image_from_path(image, path):
    if not path or not os.path.isfile(path):
        image.props.icon_name = 'unknown'
    else:
        try:
            width, height = image.get_size_request()
            if -1 in (width, height):
                width, height = 64, 64
            pixbuf = pixbuf_from_file_scaled_down(path, width, height)
            image.set_from_pixbuf(pixbuf)
            return True
        except GLib.Error:
            image.props.icon_name = 'file-broken'
    return False


def check_path_accessibility(path, file=True, executable=False):
    """Return None  if file is readable by greeter and error message otherwise"""

    # LP: #1709864, Support gtk-3.* themes
    if "gtk-3.*" in path:
        for x in range(0, 40):
            if os.path.exists(path.replace("gtk-3.*", "gtk-3.%i" % x)):
                path = path.replace("gtk-3.*", "gtk-3.%i" % x)
                return check_path_accessibility(path, file, executable)

    if not os.path.exists(path):
        return _('File not found: {path}').format(path=path)

    try:
        uid, gids = check_path_accessibility.id_cached_data
    except AttributeError:
        files = glob.glob('/etc/lightdm/lightdm.d/*.conf')
        files += ['/etc/lightdm/lightdm.conf']
        config = configparser.RawConfigParser(strict=False)
        config.read(files)
        username = config.get('LightDM', 'greeter-user', fallback='lightdm')

        pw = pwd.getpwnam(username)
        uid = pw.pw_uid
        gids = set(os.getgrouplist(username, pw.pw_gid))
        check_path_accessibility.id_cached_data = uid, gids

    parts = os.path.normpath(path).split(os.path.sep)
    if not parts[0]:
        parts[0] = os.path.sep

    def check(p):
        try:
            st = os.stat(p)
        except OSError as e:
            return _('Failed to check permissions: {error}'.format(error=e.strerror))

        if stat.S_ISDIR(st.st_mode) and not stat.S_IREAD:
            return _('Directory is not readable: {path}'.format(path=p))
        if st.st_uid == uid:
            return not (st.st_mode & stat.S_IRUSR) and \
                _('LightDM does not have permissions to read path: {path}'.format(path=p))
        if st.st_gid in gids:
            return not (st.st_mode & stat.S_IRGRP) and \
                _('LightDM does not have permissions to read path: {path}'.format(path=p))
        return not (st.st_mode & stat.S_IROTH) and \
            _('LightDM does not have permissions to read path: {path}'.format(path=p))

    errors = (check(p) for p in accumulate(parts, os.path.join))
    error = next((error for error in errors if error), None)

    if not error and file and not os.path.isfile(path):
        return _('Path is not a regular file: {path}'.format(path=path))

    if not error and executable:
        st = os.stat(path)
        if st.st_uid == uid:
            if not st.st_mode & stat.S_IXUSR:
                return _('LightDM does not have permissions to execute file: {path}'
                         .format(path=path))
        elif st.st_gid in gids:
            if not st.st_mode & stat.S_IXGRP:
                return _('LightDM does not have permissions to execute file: {path}'
                         .format(path=path))
        elif not st.st_mode & stat.S_IXOTH:
            return _('LightDM does not have permissions to execute file: {path}'.format(path=path))

    return error


def get_markup_error(markup):
    try:
        Pango.parse_markup(markup, -1, '\0')
    except GLib.Error as e:
        return e.message
    return None


def clamp(v, a, b):
    if v < a:
        return a
    if v > b:
        return b
    return v


class DefaultValueDict(defaultdict):

    def __init__(self, *items, default=None, factory=None, source=None):
        super().__init__(None, source or items)
        self._value = default
        self._factory = factory

    def __missing__(self, key):
        return self._factory(key) if self._factory else self._value


class SimpleEnumMeta(type):

    @classmethod
    def __prepare__(mcs, *args, **kwargs):
        return OrderedDict()

    def __new__(self, cls, bases, classdict):
        obj = super().__new__(self, cls, bases, classdict)
        obj._dict = OrderedDict((k, v)
                                for k, v in classdict.items() if obj._accept_member_(k, v))
        obj._tuple_type = namedtuple(obj.__class__.__name__ + 'Tuple', obj._dict.keys())
        keys = list(obj._dict.keys())
        for i in range(len(keys)):
            if obj._dict[keys[i]] is ():
                v = 0 if i == 0 else obj._dict[keys[i - 1]] + 1
                setattr(obj, keys[i], v)
                obj._dict[keys[i]] = v
        return obj

    def __contains__(self, value):
        return value in self._dict.values()

    def __iter__(self):
        return iter(self._dict.values())

    def _make(self, *args, **kwargs):
        return self._tuple_type._make(self._imake(*args, **kwargs))

    def _imake(self, *args, **kwargs):
        if args:
            return args
        elif kwargs:
            return (kwargs.get(k, v) for k, v in self._dict.items())
        else:
            return self._dict.values()


class SimpleEnum(metaclass=SimpleEnumMeta):
    _dict = None

    def __init__(self, *args, **kwargs):
        if kwargs:
            self.__dict__.update(kwargs)
        else:
            self.__dict__.update((k, args[i]) for i, k in enumerate(self._dict))

    def __iter__(self):
        return (self.__dict__[k] for k in self._dict)

    def __repr__(self):
        return repr(tuple((k, self.__dict__[k]) for k in self._dict))

    @classmethod
    def _accept_member_(cls, name, value):
        return not name.startswith('_') and not name.endswith('_')


class WidgetsEnum(SimpleEnum):

    def __init__(self, wrapper=None, builder=None):
        getter = wrapper.__getitem__ if wrapper else builder.get_object
        for k, v in self._dict.items():
            if isinstance(v, type) and issubclass(v, WidgetsEnum):
                self.__dict__[k] = v(WidgetsWrapper(wrapper or builder, k))
            else:
                self.__dict__[k] = getter(v or k)


class WidgetsWrapper:
    _builder = None
    _prefixes = None

    def __init__(self, source, *prefixes):
        if source is None:
            return
        if isinstance(source, Gtk.Builder):
            self._builder = source
            self._prefixes = tuple(prefixes)
        elif isinstance(source, WidgetsWrapper):
            self._builder = source._builder
            self._prefixes = source._prefixes + tuple(prefixes)
        else:
            raise TypeError(source)

    def __getitem__(self, args):
        if not self._builder:
            return None
        if not isinstance(args, tuple):
            args = (args,)
        return self._builder.get_object('_'.join(chain(self._prefixes, args)))

    @property
    def path(self):
        return '_'.join(self._prefixes) if self._prefixes else ''


class TreeStoreDataWrapper(GObject.Object):

    def __init__(self, data):
        super().__init__()
        self.data = data


class SimpleDictWrapper:

    def __init__(self, getter=None, setter=None, add=None, deleter=None, itergetter=None):
        self._getter = getter
        self._setter = setter
        self._deleter = deleter
        self._itergetter = itergetter
        self._add = add

    def __getitem__(self, key):
        return self._getter(key)

    def __setitem__(self, key, value):
        return self._setter(key, value)

    def __delitem__(self, key):
        return self._deleter(key)

    def __iter__(self):
        return self._itergetter()

    def add(self, value):
        return self._add(value)



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2014 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject

from lightdm_gtk_greeter_settings.helpers import WidgetsWrapper
from lightdm_gtk_greeter_settings.OptionEntry import BaseEntry
from lightdm_gtk_greeter_settings import helpers


__all__ = [
    'BaseGroup',
    'OneToManyEntryAdapter',
    'SimpleGroup']


# Broken solution - too complex
class BaseGroup(GObject.GObject):

    def __init__(self, widgets):
        super().__init__()
        self.__entries_wrapper = helpers.SimpleDictWrapper(self._get_entry)
        self.__defaults_wrapper = helpers.SimpleDictWrapper(self._get_default)

    def read(self, config):
        '''Read group content from specified GreeterConfig object'''
        raise NotImplementedError(self.__class__)

    def write(self, config, is_changed=None):
        '''Writes content of this group to specified GreeterConfig object'''
        raise NotImplementedError(self.__class__)

    def clear(self):
        '''Removes all entries'''
        raise NotImplementedError(self.__class__)

    @property
    def entries(self):
        '''entries["key"] - key => Entry mapping. Read only.'''
        return self.__entries_wrapper

    @property
    def defaults(self):
        '''defaults["key"] - default value for "key" entry. Read only.'''
        return self.__defaults_wrapper

    def _get_entry(self, key):
        raise NotImplementedError(self.__class__)

    def _get_default(self, key):
        raise NotImplementedError(self.__class__)

    @GObject.Signal
    def entry_added(self, source: object, entry: BaseEntry, key: str):
        '''New entry has been added to this group'''
        pass

    @GObject.Signal
    def entry_removed(self, source: object, entry: BaseEntry, key: str):
        '''Entry has been removed from this group'''
        pass


class SimpleGroup(BaseGroup):

    def __init__(self, name, widgets, options=None):
        super().__init__(widgets)
        self._name = name
        self._options_to_init = options
        self._widgets = WidgetsWrapper(widgets)
        self._entries = {}
        self._defaults = {}

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def options(self):
        pass

    @options.setter
    def options(self, options):
        self.clear()

        for key, (klass, default) in options.items():
            entry = klass(WidgetsWrapper(self._widgets, key))
            if default is not None:
                entry.value = default
            self._entries[key] = entry
            self._defaults[key] = default
            self.entry_added.emit(self, entry, key)

    def read(self, config):
        if self._options_to_init is not None:
            self.options = self._options_to_init
            self._options_to_init = None

        for key, entry in self._entries.items():
            value = config[self._name, key]
            entry.value = value if value is not None else self._defaults[key]
            entry.enabled = value is not None

    def write(self, config, is_changed=None):
        for key, entry in self._entries.items():
            if is_changed and not is_changed(entry):
                continue
            config[self._name, key] = entry.value if entry.enabled else None, self._get_default(key)

    def clear(self):
        if not self._entries:
            return
        for key, entry in self._entries.items():
            self.entry_removed.emit(self, entry, key)
        self._entries.clear()

    def _get_entry(self, key):
        return self._entries.get(key)

    def _get_default(self, key):
        return self._defaults.get(key)


class OneToManyEntryAdapter:

    class Error(Exception):
        pass

    class InvalidEntry(Error):
        pass

    class WrongAdapter(Error):
        pass

    class NoBaseEntryError(Error):
        pass

    class EntryWrapper(BaseEntry):

        def __init__(self, adapter):
            super().__init__(helpers.WidgetsWrapper(None))
            self._adapter = adapter
            self._value = None
            self._error = None
            self._enabled = False

        def _get_value(self):
            return self._value

        def _set_value(self, value):
            self._value = value
            if self._adapter._active == self and self._adapter._base_entry:
                self._adapter._base_entry._set_value(value)

        def _get_error(self):
            return self._error

        def _set_error(self, text):
            self._error = text
            if self._adapter._active == self and self._adapter._base_entry:
                self._adapter._base_entry._set_error(text)

        def _get_enabled(self):
            return self._enabled

        def _set_enabled(self, value):
            self._enabled = value
            if self._adapter._active == self and self._adapter._base_entry:
                self._adapter._base_entry._set_enabled(value)

    def __init__(self, entry=None):
        self._base_entry = None
        self._active = None

        self._on_changed_id = None
        self._on_menu_id = None

        self.base_entry = entry

    @property
    def base_entry(self):
        return self._base_entry

    @base_entry.setter
    def base_entry(self, entry):
        if self._base_entry:
            self._base_entry.disconnect(self._on_changed_id)
            self._base_entry.disconnect(self._on_menu_id)
            self._on_changed_id = None
            self._on_menu_id = None
        self._base_entry = entry
        if entry:
            self._on_changed_id = entry.changed.connect(self._on_changed)
            self._on_menu_id = entry.show_menu.connect(self._on_show_menu)

    def new_entry(self, widgets=None):
        return OneToManyEntryAdapter.EntryWrapper(self)

    def activate(self, entry):
        if not isinstance(entry, OneToManyEntryAdapter.EntryWrapper):
            raise OneToManyEntryAdapter.InvalidEntry()
        if not entry._adapter == self:
            raise OneToManyEntryAdapter.WrongAdapter()
        if not self._base_entry:
            raise OneToManyEntryAdapter.NoBaseEntryError()

        self._active = entry
        with self._base_entry.handler_block(self._on_changed_id):
            self._base_entry._set_value(entry._value)
            self._base_entry._set_enabled(entry._enabled)
            self._base_entry._set_error(entry._error)

    def _on_changed(self, entry):
        if self._active:
            self._active._enabled = entry._get_enabled()
            self._active._value = entry._get_value()
            self._active._emit_changed()

    def _on_show_menu(self, base_entry):
        if self._active:
            self._active.show_menu.emit()



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2014 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.


def main():
    import argparse
    import locale
    import os

    from gi.repository import Gtk
    from lightdm_gtk_greeter_settings import helpers
    from lightdm_gtk_greeter_settings import GtkGreeterSettingsWindow
    from lightdm_gtk_greeter_settings.GtkGreeterSettingsWindow import WindowMode

    locale.textdomain('lightdm-gtk-greeter-settings')

    parser = argparse.ArgumentParser(description='LightDM Gtk+ Greeter settings editor')
    parser.add_argument('-s', '--socket-id', action='store', help='Settings manager socket')
    parser.add_argument('--use-gtk-header', action='store_const', const=True,
                        help='Use GtkHeaderBar')
    parser.add_argument('--test-socket', action='store_const', const=True)
    args = parser.parse_args()

    if args.test_socket:
        w = Gtk.Window()
        w.props.title = 'Testing embedded mode'
        socket = Gtk.Socket.new()
        w.add(socket)
        w.show_all()
        args.socket_id = socket.get_id()

    try:
        socket_id = int(args.socket_id or '')
    except ValueError:
        socket_id = None

    if socket_id:
        window = GtkGreeterSettingsWindow.GtkGreeterSettingsWindow(WindowMode.Embedded)
        plug = Gtk.Plug.new(socket_id)
        plug.connect('delete-event', Gtk.main_quit)
        plug.show()
        content = window.builder.get_object('content_box')
        content.reparent(plug)
        Gtk.main()
    else:
        if args.use_gtk_header is None:
            args.use_gtk_header = helpers.string2bool(os.getenv('GTK_CSD'), False)
        elif args.use_gtk_header:
            os.putenv('GTK_CSD', '1')

        window_mode = WindowMode.Default
        if args.use_gtk_header:
            window_mode = WindowMode.GtkHeader

        window = GtkGreeterSettingsWindow.GtkGreeterSettingsWindow(mode=window_mode)
        window.show()
        Gtk.main()


if __name__ == '__main__':
    import sys
    import os
    sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
    main()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2015 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.


import os

from gi.repository import Gtk

from lightdm_gtk_greeter_settings.IconChooserDialog import IconChooserDialog
from lightdm_gtk_greeter_settings.OptionEntry import BaseEntry
from lightdm_gtk_greeter_settings.helpers import (
    C_,
    get_data_path,
    set_image_from_path,
    SimpleEnum)


__all__ = ['IconEntry']


class IconEntry(BaseEntry):

    class Item(SimpleEnum):
        priority = 0
        # (value, just_label) => (label, tooltip)
        update = None
        # (old_value) => str or None
        ask = None
        # Associated menu item
        menuitem = None

    def __init__(self, widgets):
        super().__init__(widgets)
        self._value = None
        self._image = widgets['image']
        self._button = widgets['button']
        self._button.props.popup = Gtk.Menu()
        self._icon_dialog = None
        self._path_dialog = None
        self._current_item = None

        self._add_controlled_by_state_widget(self._button)

        self._items = []
        for priority, (update, ask) in self._get_items():
            item = self.Item(priority=priority, update=update, ask=ask)
            item.menuitem = Gtk.MenuItem()
            item.menuitem.props.visible = True
            item.menuitem.props.label = item.update(None, True)[0]
            item.menuitem.connect('activate', self._on_item_clicked, item)
            self._button.props.popup.append(item.menuitem)
            self._items.append(item)

        self._items.sort(key=lambda i: i.priority)

    def _get_value(self):
        return self._value

    def _set_value(self, value):
        applied_item = None
        tooltip = None
        for item in self._items:
            if applied_item:
                label, __ = item.update(None, True)
            else:
                label, tooltip = item.update(value, False)
                if tooltip:
                    applied_item = item
            item.menuitem.get_child().set_markup(label)

        if not applied_item:
            tooltip = C_('option-entry|icon', 'Unrecognized value: {value}').format(value=value)
        self._button.set_tooltip_markup(tooltip)

        self._value = value
        self._current_item = applied_item

        self._emit_changed()

    def _get_items(self):
        return ((0, (self._update_icon, self._ask_icon)),
                (100, (self._update_image, self._ask_image)))

    def _on_item_clicked(self, menuitem, item):
        value = item.ask(self._value if item == self._current_item else None)
        if value is not None:
            self._set_value(value)

    def _update_icon(self, value, just_label=False):
        if just_label or value is None or not value.startswith('#'):
            return C_('option-entry|icon', 'Select icon name...'), None
        name = value[1:]
        label = C_('option-entry|icon', '<b>Icon: {icon}</b>').format(icon=name)
        tooltip = label
        self._image.set_from_icon_name(name, Gtk.IconSize.DIALOG)
        return label, tooltip

    def _ask_icon(self, oldvalue):
        if not self._icon_dialog:
            self._icon_dialog = IconChooserDialog()
            self._icon_dialog.props.transient_for = self._image.get_toplevel()
        if oldvalue:
            self._icon_dialog.select_icon(oldvalue[1:])

        value = None
        if self._icon_dialog.run() == Gtk.ResponseType.OK:
            value = '#' + self._icon_dialog.get_selected_icon()
        self._icon_dialog.hide()
        return value

    def _update_image(self, value, just_label=False):
        if just_label or value is None:
            return C_('option-entry|icon', 'Select file...'), None

        if set_image_from_path(self._image, value):
            label = C_('option-entry|icon', '<b>File: {path}</b>')
        else:
            label = C_('option-entry|icon', '<b>File: {path}</b> (failed to load)')

        return (label.format(path=os.path.basename(value)),
                label.format(path=value))

    def _ask_image(self, oldvalue):
        if not self._path_dialog:
            builder = Gtk.Builder()
            builder.add_from_file(get_data_path('ImageChooserDialog.ui'))

            self._path_dialog = builder.get_object('dialog')
            self._path_dialog.props.transient_for = self._image.get_toplevel()
            self._path_dialog.connect('update-preview', self._on_update_path_preview)

            preview_size = self._image.props.pixel_size
            preview = self._path_dialog.props.preview_widget
            preview.props.pixel_size = preview_size
            preview.set_size_request(preview_size, preview_size)

        if oldvalue is not None:
            self._path_dialog.select_filename(self._value)

        value = None
        if self._path_dialog.run() == Gtk.ResponseType.OK:
            value = self._path_dialog.get_filename()
        self._path_dialog.hide()
        return value

    def _on_update_path_preview(self, chooser):
        set_image_from_path(chooser.props.preview_widget, chooser.get_filename())



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2014 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.


import operator
import platform
import string
from copy import deepcopy

from gi.repository import (
    Gtk,
    Gdk)
from gi.repository.GObject import markup_escape_text as escape_markup

from lightdm_gtk_greeter_settings.helpers import (
    C_,
    get_greeter_version,
    get_markup_error,
    SimpleEnum,
    TreeStoreDataWrapper)
from lightdm_gtk_greeter_settings.OptionEntry import BaseEntry


__all__ = ['BuiltInIndicators',
           'EmptyIndicators',
           'Indicators',
           'IndicatorsEntry',
           'Option',
           'SingleIndicators']


class Indicators(SimpleEnum):
    External = '~external'
    Spacer = '~spacer'
    Separator = '~separator'
    Text = '~text'
    Host = '~host'
    Clock = '~clock'
    Layout = '~layout'
    Session = '~session'
    Language = '~language'
    A11y = '~a11y'
    Power = '~power'


# Valid builtin indicators
BuiltInIndicators = set(Indicators) - {Indicators.External}

# Special indicators
EmptyIndicators = {Indicators.Spacer, Indicators.Separator}

# These indicators can have only one instance
SingleIndicators = set(Indicators) - {Indicators.External, Indicators.Text,
                                      Indicators.Spacer, Indicators.Separator}

# Valid options names


class Option(SimpleEnum):
    # Common
    Name = 'name'
    Layout = 'layout'
    Text = 'text'
    Image = 'image'
    FallbackImage = 'fallback-image'
    Tooltip = 'tooltip'
    Markup = 'markup'
    Expand = 'expand'
    Align = 'align'
    # External
    Path = 'path'
    # Power
    HideDisabled = 'hide-disabled'


class Layout(SimpleEnum):
    Empty = ''
    Text = 'text'
    Image = 'image'
    ImageText = 'image-text'
    TextImage = 'text-image'

    @classmethod
    def _to_set(cls, value):
        return LayoutSet._str2set.get(value, set())


class LayoutSet(SimpleEnum):
    Text = 'text'
    Image = 'image'
    Reversed = 'reversed'

    _str2set = {
        Layout.Empty: set(),
        Layout.Text: {Text},
        Layout.Image: {Image},
        Layout.ImageText: {Text, Image},
        Layout.TextImage: {Text, Image, Reversed}}

    @classmethod
    def _to_string(cls, value):
        return next((k for k, v in cls._str2set.items() if v == value), '')


class Row(SimpleEnum):
    Name = ()
    Tooltip = ()
    HasState = ()
    State = ()
    Options = ()
    Icon = ()
    Markup = ()


class IndicatorsEntry(BaseEntry):
    # Readable names for indicators
    Names = {
        Indicators.External:    C_('option-entry|indicators|name', 'External library/service'),
        Indicators.Spacer:      C_('option-entry|indicators|name', 'Spacer'),
        Indicators.Separator:   C_('option-entry|indicators|name', 'Separator'),
        Indicators.Text:        C_('option-entry|indicators|name', 'Text'),
        Indicators.Clock:       C_('option-entry|indicators|name', 'Clock'),
        Indicators.Host:        C_('option-entry|indicators|name', 'Host name'),
        Indicators.Layout:      C_('option-entry|indicators|name', 'Keyboard layout'),
        Indicators.Session:     C_('option-entry|indicators|name', 'Sessions menu'),
        Indicators.Language:    C_('option-entry|indicators|name', 'Languages menu'),
        Indicators.A11y:        C_('option-entry|indicators|name', 'Accessibility menu'),
        Indicators.Power:       C_('option-entry|indicators|name', 'Power menu')}
    # Default icons for indicators to display in treeview
    Icons = {
        Indicators.A11y:        'preferences-desktop-accessibility',
        Indicators.Session:     'document-properties',
        Indicators.Power:       'system-shutdown'}
    Tooltips = {
        Indicators.Spacer:      C_('option-entry|indicators|tooltip', 'Spacer'),
        Indicators.Separator:   C_('option-entry|indicators|tooltip', 'Separator'),
        Indicators.Text:        C_('option-entry|indicators|tooltip', 'Custom text or/and image'),
        Indicators.Host:        C_('option-entry|indicators|tooltip', 'Host name'),
        Indicators.Clock:       C_('option-entry|indicators|tooltip', 'Clock'),
        Indicators.Layout:      C_('option-entry|indicators|tooltip', 'Layout indicator'),
        Indicators.Session:     C_('option-entry|indicators|tooltip',
                                   'Sessions menu (xfce, unity, gnome etc.)'),
        Indicators.Language:    C_('option-entry|indicators|tooltip', 'Languages menu'),
        Indicators.A11y:        C_('option-entry|indicators|tooltip', 'Accessibility menu'),
        Indicators.Power:       C_('option-entry|indicators|tooltip', 'Power menu')}
    # Default options for indicators
    DefaultOptions = {
        Indicators.External:    {Option.Text: None, Option.Image: None},
        Indicators.Spacer:      {Option.Layout: set()},
        Indicators.Separator:   {Option.Layout: set()},
        Indicators.Text:        {Option.Layout: {LayoutSet.Text}, Option.Text: None},
        Indicators.Host:        {Option.Layout: {LayoutSet.Text}, Option.Text: None},
        Indicators.Clock:       {Option.Layout: {LayoutSet.Text}, Option.Text: None},
        Indicators.Layout:      {Option.Layout: {LayoutSet.Text}, Option.Text: None},
        Indicators.Session:     {Option.Layout: {LayoutSet.Text, LayoutSet.Image},
                                 Option.Text: None, Option.Image: None},
        Indicators.Language:    {Option.Layout: {LayoutSet.Text}, Option.Text: None},
        Indicators.A11y:        {Option.Layout: {LayoutSet.Image}, Option.Image: None},
        Indicators.Power:       {Option.Layout: {LayoutSet.Image}, Option.Image: None}}

    def __init__(self, widgets):
        super().__init__(widgets)

        if get_greeter_version() < 0x020100:
            self._get_value = self._get_value_19
            self._on_button_release = self._on_button_release_19

        for k, v in self.DefaultOptions.items():
            v[Option.Name] = k

        self._treeview = widgets['treeview']
        self._selection = widgets['selection']
        self._state_renderer = widgets['state_renderer']
        self._state_column = widgets['state_column']
        self._add = widgets['add']
        self._remove = widgets['remove']
        self._up = widgets['up']
        self._down = widgets['down']
        self._tools = widgets['tools']
        self._model = widgets['model']
        self._widgets_to_disable = [self._treeview, widgets['toolbar']]
        self._properties_dialog = None
        self._row_menu = None
        self._tools_menu = None
        self._show_unused = False

        self._treeview.connect('key-press-event', self._on_key_press)
        self._treeview.connect('row-activated', self._on_row_activated)
        self._treeview.connect('button-release-event', self._on_button_release)
        self._selection.connect('changed', self._on_selection_changed)
        self._state_renderer.connect('toggled', self._on_state_toggled)

        self._add.connect('clicked', self._on_add_clicked)
        self._remove.connect('clicked', self._on_remove_clicked)
        self._up.connect('clicked', self._on_up_clicked)
        self._down.connect('clicked', self._on_down_clicked)
        self._tools.connect('clicked', self._on_tools_clicked)

        self._on_row_changed_id = self._model.connect('row-changed', self._on_model_changed)
        self._on_row_deleted_id = self._model.connect('row-deleted', self._on_model_changed)
        self._on_row_inserted_id = self._model.connect('row-inserted', self._on_model_row_inserted)
        self._on_rows_reordered_id = self._model.connect('rows-reordered', self._on_model_changed)

    def _on_model_changed(self, *unused):
        self._emit_changed()

    def _on_model_row_inserted(self, model, path, rowiter):
        # Do not emit 'changed' for uninitialized row (dragging rows)
        # It can cause calling get_value() for model with invalid values
        if model[rowiter][Row.Name] is not None:
            self._emit_changed()

    def _get_value(self):
        def fix_token(s):
            s = s.replace('"', r'\"')
            if any(c in s for c in string.whitespace):
                s = '"' + s + '"'
            return s

        items = []
        for row in self._model:
            if row[Row.HasState] and not row[Row.State]:
                continue

            options = deepcopy(row[Row.Options].data)
            name = options.pop(Option.Name)
            defaults = deepcopy(self.DefaultOptions[name])

            # text, image, layout=image-text -> text, image
            if options.get(Option.Layout) == {LayoutSet.Text, LayoutSet.Image}:
                del options[Option.Layout]

            for k in defaults.keys() & options.keys():
                if defaults[k] == options[k]:
                    del options[k]

            if Option.Layout in options:
                layout = options[Option.Layout]
                options[Option.Layout] = LayoutSet._to_string(layout)
                # text, layout=text -> layout=text
                if LayoutSet.Text in layout and options.get(Option.Text, self) is None:
                    del options[Option.Text]
                if LayoutSet.Image in layout and options.get(Option.Image, self) is None:
                    del options[Option.Image]

            # name=~text, text=value -> ~~value
            if name == Indicators.Text:
                name = '~~' + (options.pop(Option.Text, None) or '')
            elif name == Indicators.External:
                name = options.pop(Option.Path, None) or ''

            if not options:
                items.append(fix_token(name))
            else:
                values = (fix_token(k) + '=' + fix_token(v) if v else fix_token(k)
                          for k, v in sorted(options.items(), key=operator.itemgetter(0)))
                items.append(fix_token(name) + ': ' + ', '.join(values))
        return '; '.join(items)

    def _get_value_19(self):
        items = []
        for row in self._model:
            if row[Row.HasState] and not row[Row.State]:
                continue

            options = deepcopy(row[Row.Options].data)
            name = options.pop(Option.Name)

            # name=~text, text=value -> ~~value
            if name == Indicators.Text:
                name = '~~' + (options.pop(Option.Text, None) or '')
            elif name == Indicators.External:
                name = options.pop(Option.Path, None) or ''

            items.append(name)
        return ';'.join(items)

    def _set_value(self, value):
        with self._model.handler_block(self._on_row_deleted_id):
            self._model.clear()

        for options in self._read_options_string(value):
            name = options[Option.Name]

            if name.startswith('~~'):
                options.setdefault(Option.Text, name[2:])
                options[Option.Name] = Indicators.Text
                name = Indicators.Text
            elif name not in BuiltInIndicators:
                options.setdefault(Option.Path, name)
                options[Option.Name] = Indicators.External
                name = Indicators.External

            defaults = deepcopy(self.DefaultOptions[name])

            if Option.Markup in options:
                markup = options[Option.Markup]
                if markup is not None:
                    options[Option.Text] = markup
                options[Option.Markup] = None

            if Option.Layout in options:
                options[Option.Layout] = Layout._to_set(options[Option.Layout])
            else:
                options[Option.Layout] = defaults.get(Option.Layout) or set()

            if Option.Text in options:
                options[Option.Layout].add(LayoutSet.Text)
            elif LayoutSet.Text in options[Option.Layout]:
                options.setdefault(Option.Text, None)
            else:
                defaults.pop(Option.Text, None)

            if Option.Image in options:
                options[Option.Layout].add(LayoutSet.Image)
            elif LayoutSet.Image in options[Option.Layout]:
                options.setdefault(Option.Image, None)
            else:
                defaults.pop(Option.Image, None)

            options.update((k, defaults[k])
                           for k in defaults.keys() - options.keys())

            with self._model.handler_block(self._on_row_changed_id), \
                    self._model.handler_block(self._on_row_inserted_id):
                self._set_row(None, options, select=False)

        if self._show_unused:
            self._tools_show_unused_toggled()

        self._selection.select_path(0)
        self._on_model_changed()

    def _read_options_string(self, s):
        while s:
            name, s = self._next_string_token(s, ':;')
            options = {Option.Name: name}

            if s.startswith(':'):
                while s:
                    option, s = self._next_string_token(s[1:], '=,;')
                    if s.startswith('='):
                        value, s = self._next_string_token(s[1:], ',;')
                    else:
                        value = None
                    options[option] = value
                    if not s.startswith(','):
                        break

            yield options
            s = s[1:]

    def _next_string_token(self, s, delimiters):
        token = []
        quoted = False

        for last, c in enumerate(s):
            if not c.isspace():
                break

        # Parsing quotes
        for i, c in enumerate(s[last:], last):
            if c == '"':
                if i > last and s[i - 1] == '\\':
                    token.append(s[last:i - 1])
                    token.append('"')
                else:
                    token.append(s[last:i])
                    quoted = not quoted
                last = i + 1
            elif not quoted and c in delimiters:
                break

        if quoted:
            return '', ''

        if last != i or last == 0:
            token.append(s[last: i if c in delimiters else i + 1].rstrip())

        return ''.join(token) if token else None, s[i:]

    def _remove_selection(self):
        model, rowiter = self._selection.get_selected()
        if rowiter:
            if self._show_unused and model[rowiter][Row.HasState]:
                model[rowiter][Row.State] = False
            else:
                model.remove(rowiter)
            self._on_selection_changed()

    def _move_selection(self, move_up):
        model, rowiter = self._selection.get_selected()
        if rowiter:
            next_iter = model.iter_previous(
                rowiter) if move_up else model.iter_next(rowiter)
            if self._show_unused and \
               (model[rowiter][Row.HasState] and not model[rowiter][Row.State] or
                    model[next_iter][Row.HasState] and not model[next_iter][Row.State]):
                with self._model.handler_block(self._on_rows_reordered_id):
                    model.swap(rowiter, next_iter)
            else:
                model.swap(rowiter, next_iter)
            self._on_selection_changed()

    def _create_row_tuple(self, options):
        name = options[Option.Name]
        error = None

        text = options.get(Option.Text)
        if Option.Text in options:
            if text is not None:
                if Option.Markup in options:
                    error = get_markup_error(text)
                    if error:
                        text = '<i>{text}</i>'.format(text=escape_markup(text))
                else:
                    text = escape_markup(text)
                text = '"' + text + '"'
            elif name == Indicators.Host:
                text = escape_markup(platform.node())

        display_name = self.Names.get(name, name)
        if name == Indicators.External:
            if options.get(Option.Path):
                title = '{name} ({value})'.format(name=escape_markup(display_name),
                                                  value=escape_markup(options[Option.Path]))
            else:
                title = escape_markup(display_name)
        else:
            title = escape_markup(display_name)

        if text:
            markup = '{name}: {text}'.format(name=title, text=text)
        else:
            markup = title

        if Option.Image in options or get_greeter_version() < 0x020100:
            icon = options.get(Option.Image)
            if icon and icon.startswith('#'):
                icon = icon[1:]
            elif icon:
                icon = 'image-x-generic'
            else:
                if name in self.Icons:
                    icon = self.Icons[name]
                elif name in BuiltInIndicators:
                    icon = 'applications-system'
                else:
                    icon = 'application-x-executable'
        else:
            icon = ''

        has_state = name in SingleIndicators

        return Row._make(Name=name,
                         Tooltip=self.Tooltips.get(name),
                         Icon=icon,
                         Markup=markup,
                         HasState=has_state, State=has_state,
                         Options=TreeStoreDataWrapper(options))

    def _set_row(self, rowiter, options, select=True):
        old_name = self._model[rowiter][Row.Name] if rowiter else None
        new_name = options.get(
            Option.Name, '') if options is not None else None
        old_is_single = old_name in SingleIndicators
        new_is_single = new_name in SingleIndicators

        if new_name == old_name:
            # The same row - just update
            pass
        elif old_is_single and new_is_single:
            old_row = next(
                (row for row in self._model if row[Row.Name] == new_name), None)
            if old_row:
                if self._show_unused:
                    # Swap current row with new_row
                    with self._model.handler_block(self._on_rows_reordered_id):
                        self._model.move_before(old_row.iter, rowiter)
                    with self._model.handler_block(self._on_row_changed_id):
                        self._model[rowiter][Row.State] = False
                    rowiter = old_row.iter
                else:
                    # Replace current row with replace_row
                    with self._model.handler_block(self._on_row_deleted_id):
                        self._model.remove(old_row.iter)
        elif old_is_single:
            if self._show_unused:
                # Uncheck old row and use new instead of it
                with self._model.handler_block(self._on_row_changed_id):
                    self._model[rowiter][Row.State] = False
                with self._model.handler_block(self._on_row_inserted_id):
                    new_iter = self._model.insert_after(rowiter)
                rowiter = new_iter
        elif new_is_single:
            old_row = next(
                (row for row in self._model if row[Row.Name] == new_name), None)
            if old_row:
                with self._model.handler_block(self._on_row_deleted_id):
                    self._model.remove(old_row.iter)

        if rowiter and options:
            with self._model.handler_block(self._on_row_changed_id):
                self._model[rowiter] = self._create_row_tuple(options)
            self._model.row_changed(self._model.get_path(rowiter), rowiter)
        elif options:
            rowiter = self._model.append(self._create_row_tuple(options))

        if select and rowiter:
            self._selection.select_iter(rowiter)

        return rowiter

    def _edit_indicator(self, options, add_callback=None):
        if not self._properties_dialog:
            from lightdm_gtk_greeter_settings.IndicatorPropertiesDialog \
                import IndicatorPropertiesDialog as Dialog
            self._properties_dialog = Dialog(is_duplicate=self._is_duplicate,
                                             get_defaults=self.DefaultOptions.get,
                                             get_name=self.Names.get)
            self._properties_dialog.props.transient_for = self._treeview.get_toplevel()

        self._properties_dialog.add_callback = add_callback
        self._properties_dialog.set_indicator(options)
        if self._properties_dialog.run() == Gtk.ResponseType.OK:
            options = self._properties_dialog.get_indicator()
        else:
            options = None
        self._properties_dialog.hide()
        return options

    def _is_duplicate(self, name):
        return name in SingleIndicators and any(row[Row.Name] == name
                                                for row in self._model if row[Row.State])

    def _add_indicator(self, options):
        self._set_row(None, options)

    def _on_key_press(self, treeview, event):
        if Gdk.keyval_name(event.keyval) == 'Delete':
            self._remove_selection()
        elif Gdk.keyval_name(event.keyval) == 'F2':
            model, rowiter = self._selection.get_selected()
            treeview.row_activated(model.get_path(rowiter), None)
        else:
            return False
        return True

    def _on_row_activated(self, treeview, path, column):
        if column != self._state_column:
            options = self._edit_indicator(self._model[path][Row.Options].data)
            if options:
                self._set_row(self._model.get_iter(path), options)

    def _on_button_release(self, treeview, event):
        if event.button != 3:
            return False

        pos = treeview.get_path_at_pos(int(event.x), int(event.y))
        if not pos:
            return False

        row = self._model[pos[0]]
        if row[Row.HasState] and not row[Row.State]:
            return False

        if not self._row_menu:
            self._row_menu = Gtk.Menu()
            self._row_menu_reset = Gtk.MenuItem(C_('option-entry|indicators',
                                                   'Reset to _defaults'))
            self._row_menu_text = Gtk.CheckMenuItem(C_('option-entry|indicators',
                                                       'Display _label'))
            self._row_menu_image = Gtk.CheckMenuItem(C_('option-entry|indicators',
                                                        'Display _image'))
            self._row_menu_remove = Gtk.MenuItem(
                C_('option-entry|indicators', '_Remove'))

            self._row_menu_text_id = self._row_menu_text.connect('toggled',
                                                                 self._on_row_menu_toggled,
                                                                 Option.Text)
            self._row_menu_image_id = self._row_menu_image.connect('toggled',
                                                                   self._on_row_menu_toggled,
                                                                   Option.Image)
            self._row_menu_reset.connect(
                'activate', self._on_row_menu_reset_clicked)
            self._row_menu_remove.connect('activate', self._on_remove_clicked)

            self._row_menu.append(self._row_menu_reset)
            self._row_menu.append(self._row_menu_text)
            self._row_menu.append(self._row_menu_image)
            self._row_menu.append(Gtk.SeparatorMenuItem())
            self._row_menu.append(self._row_menu_remove)

            for item in self._row_menu:
                if type(item) is not Gtk.SeparatorMenuItem:
                    item.props.use_underline = True
                item.props.visible = True

        options = row[Row.Options].data

        with self._row_menu_text.handler_block(self._row_menu_text_id):
            self._row_menu_text.props.active = Option.Text in options
        with self._row_menu_image.handler_block(self._row_menu_image_id):
            self._row_menu_image.props.active = Option.Image in options

        editable = options[Option.Name] not in {
            Indicators.Spacer, Indicators.Separator}
        self._row_menu_reset.props.sensitive = editable
        self._row_menu_text.props.sensitive = editable
        self._row_menu_image.props.sensitive = editable

        self._row_menu.popup(None, None, None, None, 0,
                             Gtk.get_current_event_time())

        return True

    def _on_button_release_19(self, treeview, event):
        pass

    def _on_row_menu_reset_clicked(self, item):
        model, rowiter = self._selection.get_selected()
        if rowiter:
            name = model[rowiter][Row.Name]
            options = deepcopy(self.DefaultOptions[name])
            options[Option.Name] = name
            with model.handler_block(self._on_row_changed_id):
                model[rowiter] = self._create_row_tuple(options)
            model.row_changed(model.get_path(rowiter), rowiter)

    def _on_row_menu_toggled(self, item, option):
        model, rowiter = self._selection.get_selected()
        options = model[rowiter][Row.Options].data
        if item.props.active:
            options.setdefault(option, None)
        else:
            options.pop(option, None)
        model[rowiter] = self._create_row_tuple(options)

    def _on_state_toggled(self, renderer, path):
        self._model[path][Row.State] = not self._model[path][Row.State]

    def _on_selection_changed(self, selection=None):
        model, rowiter = self._selection.get_selected()
        if rowiter:
            row = model[rowiter]
            self._remove.props.sensitive = not row[
                Row.HasState] or row[Row.State]
            self._down.props.sensitive = model.iter_next(rowiter)
            self._up.props.sensitive = model.iter_previous(rowiter)
            self._treeview.scroll_to_cell(model.get_path(rowiter))
        else:
            self._remove.props.sensitive = False
            self._down.props.sensitive = False
            self._up.props.sensitive = False

    def _on_add_clicked(self, button=None):
        options = self._edit_indicator(self.DefaultOptions[Indicators.External],
                                       add_callback=self._add_indicator)
        if options:
            self._set_row(None, options, select=True)

    def _on_remove_clicked(self, button=None):
        self._remove_selection()

    def _on_up_clicked(self, button=None):
        self._move_selection(move_up=True)

    def _on_down_clicked(self, button=None):
        self._move_selection(move_up=False)

    def _on_tools_clicked(self, button=None):
        if not self._tools_menu:
            self._tools_menu = Gtk.Menu()
            self._tools_menu.attach_to_widget(self._tools)

            unused_item = Gtk.CheckMenuItem(C_('option-entry|indicators', 'Show unused items'))
            unused_item.connect('toggled', self._tools_show_unused_toggled)
            self._tools_menu.append(unused_item)

            header_item = Gtk.MenuItem(C_('option-entry|indicators', 'Predefined templates:'))
            header_item.props.sensitive = False
            self._tools_menu.append(Gtk.SeparatorMenuItem())
            self._tools_menu.append(header_item)

            templates = (
                ('host ~ clock, language, session, power',
                 '~host;~spacer;~language;~session;~power'),
                ('host ~ clock ~ language, session, a11y, power',
                 '~host;~spacer;~clock;~spacer;~language;~session;~a11y;~power'),
                ('host, layout, clock ~ language, session, power',
                 '~host;~layout;~clock;~spacer;~language;~session;~power'))

            for title, value in templates:
                item = Gtk.MenuItem(title)
                item.connect('activate', self._on_tools_template_clicked, value)
                self._tools_menu.append(item)

            self._tools_menu.show_all()
        self._tools_menu.popup(None, None, None, None, 0,
                               Gtk.get_current_event_time())

    def _on_tools_template_clicked(self, item, value):
        self._set_value(value)

    def _tools_show_unused_toggled(self, widget=None):
        if widget:
            self._show_unused = widget.props.active
        self._state_column.props.visible = self._show_unused

        used = {row[Row.Name]: row
                for row in self._model if row[Row.Name] in SingleIndicators}
        if self._show_unused:
            for name in SingleIndicators - used.keys():
                options = deepcopy(self.DefaultOptions[name])
                options[Option.Name] = name
                with self._model.handler_block(self._on_row_changed_id),\
                        self._model.handler_block(self._on_row_inserted_id):
                    rowiter = self._set_row(None, options, select=False)
                    self._model[rowiter][Row.State] = False
        else:
            for row in used.values():
                if row[Row.HasState] and not row[Row.State]:
                    with self._model.handler_block(self._on_row_deleted_id):
                        self._model.remove(row.iter)



__all__ = ["__config_path__", "__data_directory__", "__version__"]
__config_path__ = "/etc/lightdm/lightdm-gtk-greeter.conf"
__data_directory__ = '/usr/share/lightdm-gtk-greeter-settings/'
__version__ = '1.2.2'



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2014 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.


from itertools import product

from gi.repository import (
    Gdk,
    Gtk)

from lightdm_gtk_greeter_settings.helpers import WidgetsWrapper
from lightdm_gtk_greeter_settings.OptionEntry import BaseEntry


__all__ = ['PositionEntry']


class PositionEntry(BaseEntry):

    class Dimension:

        def __init__(self, widgets, on_changed):
            self._entry = widgets['entry']
            self._percents = widgets['percents']
            self._mirror = widgets['mirror']
            self._adjustment = widgets['adjustment']
            self._on_changed = on_changed

            self._anchor = None

            self._on_percents_id = self._percents.connect('toggled', self._on_percents_toggled)
            self._on_mirror_id = self._mirror.connect('toggled', self._on_mirror_toggled)
            self._on_value_id = self._adjustment.connect('value-changed', self._on_value_changed)

        @property
        def value(self):
            return '%s%d%s,%s' % ('-' if self._mirror.props.active else '',
                                  int(self._entry.props.value),
                                  '%' if self._percents.props.active else '',
                                  self._anchor)

        @value.setter
        def value(self, s):
            if isinstance(s, tuple):
                p, percents, negative, anchor = s
            else:
                value, __, anchor = s.partition(',')

                percents = value and value[-1] == '%'
                if percents:
                    value = value[:-1]

                try:
                    p = int(value)
                except ValueError:
                    p = 0

                negative = (p < 0) or (p == 0 and value and value[0] == '-')

                if not anchor or anchor not in ('start', 'center', 'end'):
                    if negative:
                        anchor = 'end'
                    else:
                        anchor = 'start'

            self._anchor = anchor
            with self._percents.handler_block(self._on_percents_id):
                self._percents.props.active = percents
            self._adjustment.props.upper = 100 if self._percents.props.active else 10000
            with self._mirror.handler_block(self._on_mirror_id):
                self._mirror.props.active = negative
            with self._adjustment.handler_block(self._on_value_id):
                self._adjustment.props.value = -p if negative else p

        @property
        def anchor(self):
            return self._anchor

        @anchor.setter
        def anchor(self, value):
            self._anchor = value

        @property
        def negative(self):
            return self._mirror.props.active

        @negative.setter
        def negative(self, value):
            self._mirror.props.active = value

        def get_value_for_screen(self, screen: int):
            p = int(self._adjustment.props.value)

            if self._percents.props.active:
                p = screen * p / 100
            if self._mirror.props.active:
                p = screen - p

            return int(p)

        def _on_percents_toggled(self, toggle):
            self._adjustment.props.upper = 100 if toggle.props.active else 10000
            self._on_changed(self)

        def _on_mirror_toggled(self, toggle):
            self._on_changed(self)

        def _on_value_changed(self, widget):
            self._on_changed(self)

    AssumedWindowSize = 430, 240

    def __init__(self, widgets):
        super().__init__(widgets)

        self._screen_size = None
        self._last_overlay_size = None
        self._last_window_allocation = None

        self._screen_frame = widgets['screen_frame']
        self._screen_overlay = widgets['screen_overlay']
        window_box = widgets['window_box']
        grid = widgets['window_grid']

        self._motion = False

        window_box.connect('motion-notify-event', self._on_window_motion)
        window_box.connect('button-press-event', self._on_window_button_press)
        window_box.connect('button-release-event', self._on_window_button_release)

        # Creating points grid
        anchors_align = (Gtk.Align.START, Gtk.Align.CENTER, Gtk.Align.END)
        anchors = [(x, y, Gtk.RadioButton())
                   for x, y in product(enumerate(('start', 'center', 'end')), repeat=2)]

        self._anchors = {}
        self._anchors_handlers = {}
        for (left, x_anchor), (top, y_anchor), w in anchors:
            w.props.halign = anchors_align[left]
            w.props.valign = anchors_align[top]
            if w != anchors[0][-1]:
                w.props.group = anchors[0][-1]
            grid.attach(w, left, top, 1, 1)
            self._anchors[x_anchor, y_anchor] = w
            self._anchors_handlers[w] = w.connect('toggled', self._on_anchor_toggled,
                                                  x_anchor, y_anchor)

        grid.show_all()

        self._x = PositionEntry.Dimension(WidgetsWrapper(widgets, 'x'), self._on_dimension_changed)
        self._y = PositionEntry.Dimension(WidgetsWrapper(widgets, 'y'), self._on_dimension_changed)

        self._on_gdk_screen_changed()

        self._screen_overlay.connect('get-child-position',
                                     self._on_screen_overlay_get_child_position)
        self._screen_overlay.connect('screen-changed', self._on_gdk_screen_changed)

    def _get_value(self):
        x = self._x.value
        y = self._y.value
        return x + ' ' + y if x != y else x

    def _set_value(self, value):
        self._last_window_allocation = None
        if value:
            x, __, y = value.partition(' ')
            self._x.value = x
            self._y.value = y or x
            anchor = self._anchors[self._x.anchor, self._y.anchor]
            with anchor.handler_block(self._anchors_handlers[anchor]):
                anchor.props.active = True
        self._screen_overlay.queue_resize()
        self._emit_changed()

    def _get_corrected_position(self, p, screen, window, anchor):
        if anchor == 'center':
            p -= window / 2
        elif anchor == 'end':
            p -= window

        if p + window > screen:
            p = screen - window
        if p < 0:
            p = 0

        return int(p)

    def _on_dimension_changed(self, dimension):
        self._last_window_allocation = None
        self._screen_overlay.queue_resize()
        self._emit_changed()

    def _on_screen_overlay_get_child_position(self, overlay, child, allocation):
        screen = overlay.get_allocation()

        if self._last_window_allocation and \
                self._last_overlay_size == (screen.width, screen.height):
            (allocation.x, allocation.y,
             allocation.width, allocation.height) = self._last_window_allocation
            return True
        self._last_overlay_size = screen.width, screen.height

        scale = screen.width / self._screen_size[0]

        width = int(self.AssumedWindowSize[0] * scale)
        height = int(self.AssumedWindowSize[1] * scale)

        # Set desired size
        child.set_size_request(width, height)
        # And check what actually we have now
        width, height = child.size_request().width, child.size_request().height

        x = int(self._x.get_value_for_screen(self._screen_size[0]) * scale)
        y = int(self._y.get_value_for_screen(self._screen_size[1]) * scale)

        x = self._get_corrected_position(x, screen.width, width, self._x.anchor)
        y = self._get_corrected_position(y, screen.height, height, self._y.anchor)

        self._last_window_allocation = x, y, width, height
        allocation.x, allocation.y, allocation.width, allocation.height = x, y, width, height

        return True

    def _on_anchor_toggled(self, toggle, x, y):
        if not toggle.props.active:
            return
        self._x.anchor = x
        self._y.anchor = y
        self._last_window_allocation = None
        self._screen_overlay.queue_resize()
        self._emit_changed()

    def _on_gdk_screen_changed(self, widget=None, prev_screen=None):
        screen = self._screen_overlay.get_toplevel().get_screen()
        geometry = screen.get_monitor_geometry(screen.get_primary_monitor())
        self._screen_size = geometry.width, geometry.height
        self._screen_frame.props.ratio = geometry.width / geometry.height

    def _on_window_motion(self, widget, event):
        if not event.state & Gdk.ModifierType.BUTTON1_MASK:
            return False

        screen = self._screen_overlay.get_allocation()
        x, y = widget.translate_coordinates(self._screen_overlay, event.x, event.y)

        for d, p in ((self._x, int(100 * x / screen.width)),
                     (self._y, int(100 * y / screen.height))):
            if p < 0:
                p = 0
            elif p > 100:
                p = 100
            if d.negative:
                d.value = p - 100, True, True, d.anchor
            else:
                d.value = p, True, False, d.anchor

        self._motion = True
        self._last_window_allocation = None
        self._screen_overlay.queue_resize()
        return True

    def _on_window_button_press(self, widget, event):
        if event.button == 1:
            self._motion = False

    def _on_window_button_release(self, widget, event):
        if self._motion and event.button == 1:
            self._motion = False
            self._emit_changed()



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   LightDM GTK Greeter Settings
#   Copyright (C) 2014 Andrew P. <pan.pav.7c5@gmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.


import time
from locale import gettext as _

from gi.repository import (
    Gdk,
    GLib,
    GObject,
    Gtk)
from lightdm_gtk_greeter_settings.helpers import (
    C_,
    bool2string,
    string2bool,
    SimpleEnum)


__all__ = [
    'AccessibilityStatesEntry',
    'AdjustmentEntry',
    'BackgroundEntry',
    'BaseEntry',
    'BooleanEntry',
    'ChoiceEntry',
    'ClockFormatEntry',
    'InvertedBooleanEntry',
    'StringEntry',
    'StringPathEntry']


class BaseEntry(GObject.GObject):

    def __init__(self, widgets):
        super().__init__()
        self._widgets = widgets
        self.__widgets_to_disable = []

        self.__use = widgets['use']
        if self.__use:
            self.__use.connect('notify::active', self.__on_use_toggled)

        self.__error = widgets['error']

        self._add_label_widget(widgets['label_holder'])

    @property
    def value(self):
        '''Option value'''
        value = self._get_value()
        formatted = self.get.emit(value)
        return value if formatted is None else formatted

    @value.setter
    def value(self, value):
        if self.__use:
            self.__use.set_active(True)
        formatted = self.set.emit(value)
        self._set_value(value if formatted is None else formatted)

    @property
    def enabled(self):
        return self._get_enabled()

    @enabled.setter
    def enabled(self, value):
        self._set_enabled(value)

    @property
    def error(self):
        return self._get_error()

    @error.setter
    def error(self, value):
        self._set_error(value)

    @property
    def widgets(self):
        return self._widgets

    @GObject.Signal
    def changed(self):
        pass

    @GObject.Signal(flags=GObject.SIGNAL_RUN_CLEANUP)
    def get(self, value: str) -> str:
        pass

    @GObject.Signal(flags=GObject.SIGNAL_RUN_CLEANUP)
    def set(self, value: str) -> str:
        pass

    @GObject.Signal('show-menu')
    def show_menu(self):
        pass

    def __repr__(self):
        try:
            value = self._get_value()
        except NotImplemented:
            value = '<Undefined>'
        return '%s(%s:%s)' % (self.__class__.__name__, int(self.enabled),
                              value)

    def _get_value(self):
        raise NotImplementedError(self.__class__)

    def _set_value(self, value):
        raise NotImplementedError(self.__class__)

    def _get_error(self):
        if self.__error:
            return self.__error.props.tooltip_text
        return None

    def _set_error(self, text):
        if self.__error:
            self.__error.props.visible = text is not None
            self.__error.props.tooltip_text = text

    def _get_enabled(self):
        if self.__use:
            return self.__use.get_active()
        return True

    def _set_enabled(self, value):
        if self.__use:
            self.__use.set_active(value)
            for widget in self.__widgets_to_disable:
                widget.props.sensitive = value
        self._emit_changed()

    def _add_label_widget(self, *widgets):
        for widget in widgets:
            if widget:
                widget.connect('button-press-event', self.__on_label_clicked)

    def _add_controlled_by_state_widget(self, *widgets):
        self.__widgets_to_disable += widgets

    def _show_menu(self):
        self.__on_label_clicked()

    def _emit_changed(self, *unused):
        self.changed.emit()

    def __on_use_toggled(self, toggle, *unused):
        self._set_enabled(self.__use.props.active)

    def __on_label_clicked(self, widget, event):
        if event.button == 3:
            self.show_menu.emit()


class BooleanEntry(BaseEntry):

    def __init__(self, widgets):
        super().__init__(widgets)
        self._value = widgets['value']
        self._value.connect('notify::active', self._emit_changed)
        self._add_controlled_by_state_widget(self._value)

    def _get_value(self):
        return bool2string(self._value.props.active)

    def _set_value(self, value):
        self._value.props.active = string2bool(value)


class InvertedBooleanEntry(BooleanEntry):

    def __init__(self, widgets):
        super().__init__(widgets)

    def _get_value(self):
        return bool2string(not self._value.props.active)

    def _set_value(self, value):
        self._value.props.active = not string2bool(value)


class StringEntry(BaseEntry):

    def __init__(self, widgets):
        super().__init__(widgets)
        self._value = widgets['value']
        self._add_controlled_by_state_widget(self._value)
        if isinstance(self._value.props.parent, Gtk.ComboBox):
            self._add_controlled_by_state_widget(self._value.props.parent)
        self._value.connect('changed', self._emit_changed)

    def _get_value(self):
        return self._value.props.text

    def _set_value(self, value):
        self._value.props.text = value or ''


class StringPathEntry(BaseEntry):

    class Row(SimpleEnum):
        Title = ()
        Type = ()

    class ItemType(SimpleEnum):
        Select = 'select-path'
        Value = 'value'
        Separator = 'separator'

    def __init__(self, widgets):
        super().__init__(widgets)

        self._file_dialog = None
        self._combo = widgets['combo']
        self._entry = widgets['entry']
        self._filters = ()
        self._add_controlled_by_state_widget(self._combo)

        self._entry.connect('changed', self._emit_changed)
        self._combo.connect('format-entry-text', self._on_combobox_format)

        self._combo.set_row_separator_func(self._row_separator_callback, None)

    def add_filter(self, file_filter):
        if self._file_dialog:
            self._file_dialog.add_filter(file_filter)
        elif self._filters:
            self._filters.append(file_filter)
        else:
            self._filters = [file_filter]

    def _get_value(self):
        return self._entry.props.text

    def _set_value(self, value):
        self._entry.props.text = value or ''

    def _row_separator_callback(self, model, rowiter, data):
        return model[rowiter][self.Row.Type] == self.ItemType.Separator

    def _on_combobox_format(self, combobox, path):
        value = self._entry.props.text
        item_id = combobox.get_active_id()
        if item_id == self.ItemType.Select:
            if not self._file_dialog:
                self._file_dialog = Gtk.FileChooserDialog(
                    parent=self._combo.get_toplevel(),
                    buttons=(_('_OK'), Gtk.ResponseType.OK,
                             _('_Cancel'), Gtk.ResponseType.CANCEL),
                    title=C_('option|StringPathEntry', 'Select path'))
                for f in self._filters:
                    self._file_dialog.add_filter(f)
            if self._file_dialog.run() == Gtk.ResponseType.OK:
                value = self._file_dialog.get_filename()
            self._file_dialog.hide()
        elif item_id == self.ItemType.Value:
            value = combobox.props.model[path][self.Row.Title]

        combobox.set_active(-1)
        combobox.grab_focus()
        return value


class AdjustmentEntry(BaseEntry):

    def __init__(self, widgets):
        super().__init__(widgets)
        self._value = widgets['adjustment']
        self._value.connect('value-changed', self._emit_changed)
        self._view = widgets['view']
        self._add_controlled_by_state_widget(self._view)

    def _get_value(self):
        return str(self._value.props.value)

    def _set_value(self, value):
        try:
            self._value.props.value = float(value or '')
        except ValueError:
            self._value.props.value = 0


class ChoiceEntry(BaseEntry):

    def __init__(self, widgets):
        super().__init__(widgets)
        self._value = widgets['value']
        self._value.connect('changed', self._emit_changed)
        self._add_controlled_by_state_widget(self._value)

    def _get_value(self):
        return self._value.props.active_id

    def _set_value(self, value):
        self._value.props.active_id = value or ''


class ClockFormatEntry(StringEntry):

    def __init__(self, widgets):
        super().__init__(widgets)
        self._preview = widgets['preview']
        self._value.connect('changed', self._on_changed)
        GLib.timeout_add_seconds(1, self._on_changed, self._value)

    def _on_changed(self, entry):
        self._preview.props.label = time.strftime(self._value.props.text)
        return True


class BackgroundEntry(BaseEntry):

    def __init__(self, widgets):
        super().__init__(widgets)
        self._image_choice = widgets['image_choice']
        self._color_choice = widgets['color_choice']
        self._image_value = widgets['image_value']
        self._color_value = widgets['color_value']

        self._add_controlled_by_state_widget(self._image_choice, self._color_choice,
                                             self._image_value, self._color_value)

        self._on_choice_id = self._color_choice.connect('toggled', self._on_color_choice_toggled)
        self._color_value.connect('color-set', self._on_color_set)
        self._image_value.connect('file-set', self._on_file_set)

    def _get_value(self):
        if self._image_choice.props.active:
            return self._image_value.get_filename() or ''
        else:
            r, g, b, __ = (int(0xFF * v) for v in self._color_value.props.rgba)
            return '#%02x%02x%02x' % (r, g, b)

    def _set_value(self, value):
        if value is None:
            value = ''

        rgba = Gdk.RGBA()
        if not rgba.parse(value):
            rgba = None

        with self._color_choice.handler_block(self._on_choice_id):
            self._color_choice.props.active = rgba is not None
            self._image_choice.props.active = rgba is None

        if rgba is not None:
            self._color_value.props.rgba = rgba
            self._image_value.unselect_all()
        else:
            if value:
                self._image_value.select_filename(value)
            else:
                self._image_value.unselect_all()

        self._emit_changed()

    def _on_color_choice_toggled(self, toggle):
        self._emit_changed()

    def _on_color_set(self, button):
        if not self._color_choice.props.active:
            self._color_choice.props.active = True
        else:
            self._emit_changed()

    def _on_file_set(self, button):
        if not self._image_choice.props.active:
            self._image_choice.props.active = True
        else:
            self._emit_changed()


class FontEntry(BaseEntry):

    def __init__(self, widgets):
        super().__init__(widgets)
        self._value = widgets['value']
        self._value.connect('font-set', self._emit_changed)
        self._add_controlled_by_state_widget(self._value)

    def _get_value(self):
        return self._value.get_font_name()

    def _set_value(self, value):
        self._value.props.font_name = value or ''


class AccessibilityStatesEntry(BaseEntry):

    Options = ('keyboard', 'reader', 'contrast', 'font')

    def __init__(self, widgets):
        super().__init__(widgets)

        self._states = {name: widgets[name] for name in self.Options}

        for w in self._states.values():
            w.connect('changed', self._emit_changed)

    def _get_value(self):
        states = {name: widget.props.active_id
                  for (name, widget) in self._states.items()}
        return ';'.join(state + name
                        for (name, state) in states.items() if state not in {None, '-'})

    def _set_value(self, value):
        if value:
            states = dict((v[1:], v[0]) if v[0] in ('-', '+', '~') else (v, '-')
                          for v in value.split(';') if v)
        else:
            states = {}
        for name in self.Options:
            self._states[name].props.active_id = states.get(name, '-')
